r"""
Prior observations
"""


#[

from __future__ import annotations

import numpy as _np
import math as _mt
from numbers import Real
from typing import Protocol

from ._dimensions import Dimensions

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import NoReturn, Self

#]


__all__ = (
    "MinnesotaPriorObs",
    "MeanPriorObs",
)


class PriorObs(Protocol, ):
    r"""
    """
    #[

    def get_num_obs(self, dimensions: Dimensions, ) -> int:
        ...

    def generate_y0(
        self,
        dimensions: Dimensions,
        y_std: _np.ndarray | Real = 1,
    ) -> _np.ndarray:
        ...

    def generate_y1(
        self,
        dimensions: Dimensions,
        y_std: _np.ndarray | Real = 1,
    ) -> _np.ndarray:
        ...

    def generate_x(
        self,
        dimensions: Dimensions,
        y_std: _np.ndarray | Real = 1,
    ) -> _np.ndarray:
        ...

    def generate_k(
        self,
        dimensions: Dimensions,
        y_std: _np.ndarray | Real = 1,
    ) -> _np.ndarray:
        ...

    def __iter__(self, ) -> Iterable[Self, ]:
        yield self

    def generate_lhs(
        self,
        dimensions: Dimensions,
        y_std: _np.ndarray | Real = 1,
    ) -> _np.ndarray:
        r"""
        """
        return self.generate_y0(dimensions, )

    def generate_rhs(
        self,
        dimensions: Dimensions,
        y_std: _np.ndarray | Real = 1,
    ) -> _np.ndarray:
        r"""
        """
        return _np.vstack([
            self.generate_y1(dimensions, ),
            self.generate_x(dimensions, ),
            self.generate_k(dimensions, ),
        ])

    def _populate_mu(
        self,
        mu: Real | None = None,
        mu2: Real | None = None,
    ) -> None:
        r"""
        """
        if (mu is not None) + (mu2 is not None) != 1:
            raise ValueError("Exactly one of mu and mu2 must be specified.")
        self.mu = mu if mu is not None else _mt.sqrt(mu2)

    #]


class MinnesotaPriorObs(PriorObs, ):
    r"""
    """
    #[

    def __init__(
        self,
        rho: Real | _np.ndarray = 0,
        mu: Real | None = None,
        mu2: Real | None = None,
        kappa: Real = 0,
    ) -> None:
        r"""
        """
        self.rho = rho
        self._populate_mu(mu, mu2, )
        self.kappa = kappa

    def generate_y0(
        self,
        dimensions: Dimensions,
        y_std: _np.ndarray | Real = 1,
    ) -> _np.ndarray:
        r"""
        """
        dims = Dimensions(*dimensions)
        num_endogenous, order, = dims.num_endogenous, dims.order,
        try:
            rho = _ensure_array(self.rho, num_endogenous, )
        except ValueError:
            raise ValueError("Invalid dimensions of MinnesotaPriorObs parameter rho.", )
        y_std = _ensure_array(y_std, num_endogenous, )
        mu_scaled = self.mu * y_std
        return _np.hstack([
            _np.diag(mu_scaled * rho),
            _np.zeros((num_endogenous, num_endogenous * (order - 1)), ),
        ], dtype=float, )

    def generate_y1(
        self,
        dimensions: Dimensions,
        y_std: _np.ndarray | Real = 1,
    ) -> _np.ndarray:
        r"""
        """
        dims = Dimensions(*dimensions)
        num_endogenous, order, = dims.num_endogenous, dims.order,
        y_std = _ensure_array(y_std, num_endogenous, )
        mu_scaled = self.mu * y_std
        order_multipliers = ( (i+1)**self.kappa for i in range(order) )
        y1_diag = _np.hstack([ mu_scaled * i for i in order_multipliers ], dtype=float, )
        return _np.diag(y1_diag, )

    def generate_x(
        self,
        dimensions: Dimensions,
        y_std: _np.ndarray | Real = 1,
    ) -> None:
        r"""
        """
        dims = Dimensions(*dimensions)
        num_exogenous = dims.num_exogenous
        num_obs = self.get_num_obs(dimensions, )
        shape = (num_exogenous, num_obs, )
        return _np.zeros(shape, dtype=float, )

    def generate_k(
        self,
        dimensions: Dimensions,
        y_std: _np.ndarray | Real = 1,
    ) -> None:
        r"""
        """
        dims = Dimensions(*dimensions)
        num_obs = self.get_num_obs(dimensions, )
        shape = (int(dims.has_intercept), num_obs, )
        return _np.zeros(shape, dtype=float, )

    def get_num_obs(self, dimensions: Dimensions, ) -> int:
        r"""
        """
        dims = Dimensions(*dimensions)
        num_endogenous, order, = dims.num_endogenous, dims.order,
        return num_endogenous * order

    #]


class MeanPriorObs(PriorObs, ):
    r"""
    """
    #[

    def __init__(
        self,
        mean: Real | _np.ndarray = 0,
        mu: Real | None = None,
        mu2: Real | None = None,
    ) -> None:
        r"""
        """
        self._populate_mu(mu, mu2, )
        self.mean = mean

    def generate_y0(
        self,
        dimensions: Dimensions,
        y_std: _np.ndarray | Real = 1,
    ) -> _np.ndarray:
        r"""
        """
        dims = Dimensions(*dimensions)
        num_endogenous, has_intercept, = dims.num_endogenous, dims.has_intercept,
        try:
            mean = _ensure_array(self.mean, dims.num_endogenous, )
        except ValueError:
            raise ValueError("Invalid dimensions of MeanPriorObs parameter mean.", )
        shape = (dims.num_endogenous, int(dims.has_intercept), )
        return _np.ones(shape, dtype=float, ) * mean.reshape((-1, 1)) * self.mu

    def generate_y1(
        self,
        dimensions: Dimensions,
        y_std: _np.ndarray | Real = 1,
    ) -> _np.ndarray:
        r"""
        """
        y0 = self.generate_y0(dimensions, y_std, )
        dims = Dimensions(*dimensions)
        return _np.tile(y0, (dims.order, 1, ))

    def generate_x(
        self,
        dimensions: Dimensions,
        y_std: _np.ndarray | Real = 1,
    ) -> None:
        r"""
        """
        dims = Dimensions(*dimensions)
        shape = (dims.num_exogenous, self.get_num_obs(dimensions, ), )
        return _np.zeros(shape, dtype=float, )

    def generate_k(
        self,
        dimensions: Dimensions,
        y_std: _np.ndarray | Real = 1,
    ) -> None:
        r"""
        """
        num_obs = self.get_num_obs(dimensions, )
        dims = Dimensions(*dimensions)
        shape = (int(dims.has_intercept), num_obs, )
        return _np.ones(shape, dtype=float, ) * self.mu

    def get_num_obs(self, dimensions: Dimensions, ) -> int:
        r"""
        """
        dims = Dimensions(*dimensions)
        return int(dims.has_intercept)

    #]


def _ensure_array(
    x: Real | _np.ndarray,
    num: int,
) -> _np.ndarray | NoReturn:
    r"""
    """
    #[
    shape = (num, )
    if not isinstance(x, _np.ndarray) or x.size == 1:
        x = _np.full(shape, x)
    if x.shape != shape:
        raise ValueError(f"Expected shape {shape} but got {x.shape}", )
    return x.astype(float, )
    #]


def arrays_from_prior_obs(
    prior_obs: PriorObs | Iterable[PriorObs, ] | None,
    dimensions: Dimensions,
    y_std: _np.ndarray | Real = 1,
) -> tuple[_np.ndarray, _np.ndarray, ] | None:
    r"""
    """
    #[
    if prior_obs is None:
        return None
    lhs, rhs, = zip(*(
        (i.generate_lhs(dimensions, y_std, ), i.generate_rhs(dimensions, y_std, ), )
        for i in prior_obs
    ))
    lhs = _np.hstack(lhs, )
    rhs = _np.hstack(rhs, )
    return lhs, rhs,
    #]




"""
Time series module
"""

from .main import *
from .main import __all__ as main_all

from .functions import *
from .functions import __all__ as functions_all

__all__ = []
__all__.extend(main_all)
__all__.extend(functions_all)




"""
Implement SteadyDataboxableProtocol
"""


#[

from __future__ import annotations

import numpy as _np

from .. import has_variants as _has_variants
from ..quantities import QuantityKind
from ..series.main import Series

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from collections.abc import Iterable
    from numbers import Real
    from ..dates import Period

#]


r"""
Quantities that are time series in model databoxes
"""
_TIME_SERIES_QUANTITY = (
    QuantityKind.ANY_VARIABLE
    | QuantityKind.ANY_SHOCK_OR_SHOCK_VALUE
)


#-------------------------------------------------------------------------------
# Functions to be used as methods in Simultaneous class
#-------------------------------------------------------------------------------


def inlay(klass: type, ) -> type:
    r"""
    Inlay the steady state methods in the class
    """
    #[
    klass.generate_steady_items = generate_steady_items
    return klass
    #]


def generate_steady_items(
    self,
    start: Period,
    end: Period,
    *,
    deviation: bool = False,
    unpack_singleton: bool = True,
) -> Iterable[tuple[str, Series | Real]]:
    """
    """
    num_periods = int(end - start + 1)
    shift_in_first_column = start.get_distance_from_origin()
    qid_to_name = self.create_qid_to_name()
    qid_to_description = self.create_qid_to_description()
    #
    tuple_of_arrays = tuple(
        self.create_some_array(
        variant=variant,
        deviation=deviation,
        num_columns=num_periods,
        shift_in_first_column=shift_in_first_column,
    ) for variant in self._variants)
    array = _np.stack(arrays=tuple_of_arrays, axis=2, )
    #
    num_rows = array.shape[0]
    qid_to_kind = self.create_qid_to_kind()
    for qid in qid_to_name.keys():
        name = qid_to_name[qid]
        array_slice = array[qid, :, :]
        is_time_series = qid_to_kind[qid] in _TIME_SERIES_QUANTITY
        if is_time_series:
            value = Series(
                start=start,
                values=array_slice,
                description=qid_to_description[qid],
            )
        else:
            value = _has_variants.unpack_singleton(
                array_slice[0, :].tolist(), self.is_singleton,
                unpack_singleton=unpack_singleton,
            )
        yield name, value,


#-------------------------------------------------------------------------------




"""
Dynamic nonlinear period-by-period simulator
"""


#[

from __future__ import annotations

import numpy as _np
import scipy as _sp

from .. import equations as _equations
from .. import wrongdoings as _wrongdoings
from .. import frames as _frames
from ..simultaneous import main as _simultaneous
from ..plans.simulation_plans import SimulationPlan
from ..dataslates.main import Dataslate
from ..fords import terminators as _terminators

from . import _evaluators as _evaluators
from ..stacked_time import simulators as _stacked_time_simulators

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from numbers import Real
    from typing import Callable, Any, Literal
    from ..dates import Period
    from ..frames import Frame
    from ..incidences.main import Token

#]


METHOD_NAME = "period_by_period"


def create_frames(
    model_v: Simultaneous,
    dataslate_v: Dataslate,
    plan: SimulationPlan | None,
    **kwargs,
) -> tuple[Frame, ...]:
    """
    """
    #[
    base_break_points = _frames.create_empty_base_break_points(dataslate_v, )
    base_break_points[:] = True
    base_end = dataslate_v.base_periods[-1]
    return _frames.split_into_frames_by_breakpoints(
        base_break_points, dataslate_v,
        get_simulation_end=lambda start, end: end,
    )
    #]


def simulate_initial_guess(
    *args,
    initial_guess: Literal["first_order", "data"] = "data",
    **kwargs,
) -> None:
    """
    """
    return _stacked_time_simulators.simulate_initial_guess(
        *args,
        initial_guess=initial_guess,
        **kwargs,
    )


def simulate_frame(
    model_v: Simultaneous,
    frame_ds: Dataslate,
    *,
    frame: Frame,
    **kwargs,
) -> dict[str, Any]:
    """
    """
    is_first_frame = frame_ds.base_columns[0] == frame.first
    return _stacked_time_simulators.simulate_frame(
        model_v,
        frame_ds,
        frame=frame,
        **kwargs,
        terminal="data",
        _precatch_missing=_precatch_missing_first_frame if is_first_frame else _precatch_missing,
    )


def _setup_current_period(
    plan: SimulationPlan | None,
    create_evaluator: Callable,
    current_wrt_qids: tuple[int, ...],
    current_period: Period,
    name_to_qid: dict[str, int],
    /,
) -> tuple[tuple[int, ...], PeriodEvaluator]:
    """
    """
    current_wrt_qids = tuple(current_wrt_qids)
    if plan:
        names_exogenized = plan.get_exogenized_unanticipated_in_period(current_period, )
        names_endogenized = plan.get_endogenized_unanticipated_in_period(current_period, )
        if len(names_exogenized) != len(names_endogenized):
            raise _wrongdoings.IrisPieCritical(
                f"Number of exogenized quantities {len(names_exogenized)}"
                f" does not match number of endogenized quantities {len(names_endogenized)}"
                f" in period {current_period}"
            )
        qids_exogenized = tuple(name_to_qid[name] for name in names_exogenized)
        qids_endogenized = tuple(name_to_qid[name] for name in names_endogenized)
        current_wrt_qids = tuple(sorted(set(current_wrt_qids).difference(qids_exogenized).union(qids_endogenized)))
    current_evaluator = create_evaluator(current_wrt_qids, )
    return current_wrt_qids, current_evaluator


def _initial_guess_previous_period(
    data: _np.ndarray,
    wrt_qids: tuple[int, ...],
    t: int,
) -> _np.ndarray:
    """
    """
    source_t = t - 1 if t > 0 else 0
    return data[wrt_qids, source_t]


def _initial_guess_data(
    data: _np.ndarray,
    wrt_qids: tuple[int, ...],
    t: int,
) -> _np.ndarray:
    """
    """
    return data[wrt_qids, t]


def _catch_missing(
    data: _np.ndarray,
    t: int,
    /,
    frame_ds: Dataslate,
    qid_to_name: dict[int, str],
    fallback_value: Real,
    when_missing_stream: _wrongdoings.Stream,
) -> None:
    """
    """
    missing = _np.isnan(data[:, t])
    if not missing.any():
        return
    #
    data[missing, t] = fallback_value
    #
    current_period = frame_ds.periods[t]
    shift = 0
    for qid in _np.flatnonzero(missing):
        when_missing_stream.add(
            f"{qid_to_name[qid]}[{current_period+shift}]"
            f" when simulating {current_period}"
        )


def _catch_fail(
    when_fails_stream,
    root_final: _sp.optimize.OptimizeResult,
    period: Period,
) -> None:
    """
    """
    when_fails_stream.add(
        f"Simulation failed in {period} with message: {root_final.message}"
    )


def _create_custom_header(
    vid: str,
    t: int,
    current_period: int,
) -> str:
    """
    """
    return f"[Variant {vid}][Period {current_period}]"


def _get_wrt_qids(
    model_v,
    name_to_qid: dict[str, int],
    /,
) -> tuple[int, ...]:
    """
    """
    plannable = model_v.get_simulation_plannable()
    wrt_names \
        = tuple(plannable.can_be_exogenized_unanticipated) \
        + tuple(plannable.can_be_exogenized_anticipated)
    wrt_qids = sorted(
        name_to_qid[name]
        for name in set(wrt_names)
    )
    return wrt_qids


_INITIAL_GUESS = {
    "previous_period": _initial_guess_previous_period,
    "data": _initial_guess_data,
}


_DEFAULT_SOLVER_SETTINGS = {
    "method": "hybr",
    "tol": 1e-12,
    "options": {"xtol": 1e-12, },
}


def _precatch_missing(
    data: _np.ndarray,
    wrt_spots: Iterable[Token],
) -> None:
    """
    In the second and subsequent periods, always copy the previous period's
    values for endogenous quantities
    """
    for qid, column in wrt_spots:
        if column == 0:
            continue
        data[qid, column] = data[qid, column-1]


def _precatch_missing_first_frame(
    data: _np.ndarray,
    wrt_spots: Iterable[Token],
) -> None:
    """
    In the first period, copy the previous period's values for endogenous
    quantities only if they are NaN
    """
    for qid, column in wrt_spots:
        if column == 0 or not _np.isnan(data[qid, column]):
            continue
        data[qid, column] = data[qid, column-1]




"""
Functional forms of Series methods
"""


#[

from __future__ import annotations

from ._temporal import __all__ as __all__temporal
from ._temporal import *

from ._filling import __all__ as __all__filling
from ._filling import *

from ._conversions import __all__ as __all__conversions
from ._conversions import *

from ._hp import __all__ as __all__hp
from ._hp import *

from ._extrapolate import __all__ as __all__extrapolate
from ._extrapolate import *

from ._x13 import __all__ as __all__x13
from ._x13 import *

from ._moving import __all__ as __all__moving
from ._moving import *

from ._statistics import __all__ as __all__statistics
from ._statistics import *

from ._elementwise import __all__ as __all__elementwise
from ._elementwise import *

from ._ell_one import __all__ as __all__ell_one
from ._ell_one import *

#]


__all__ = []
__all__.extend(__all__temporal)
__all__.extend(__all__filling)
__all__.extend(__all__conversions)
__all__.extend(__all__hp)
__all__.extend(__all__extrapolate)
__all__.extend(__all__x13)
__all__.extend(__all__moving)
__all__.extend(__all__statistics)
__all__.extend(__all__elementwise)
__all__.extend(__all__ell_one)





"""
"""


#[
from __future__ import annotations

from collections.abc import (Iterable, )
import re as _re
import parsimonious as _pa
#]


GRAMMAR_DEF = r"""

    keyword_prefix = "¡"
    human_prefix = "!"
    white_spaces = ~r"[ \t\n]*"
    non_keyword_prefixes = ~r"[^¡]+"

"""


GRAMMAR = _pa.grammar.Grammar(GRAMMAR_DEF)
KEYWORD_PREFIX = GRAMMAR["keyword_prefix"].literal
HUMAN_PREFIX = GRAMMAR["human_prefix"].literal


def compile_keywords_pattern(commands_list: Iterable[str]) -> _re.Pattern:
    return _re.compile(
        HUMAN_PREFIX + r"(" + "|".join(commands_list) + r")\b"
    )


def translate_keywords(commands_pattern: _re.Pattern, code: str) -> str:
    return _re.sub(commands_pattern, lambda m: KEYWORD_PREFIX + m.group(1), code)


def add_blank_lines(source: str) -> str:
    return "\n" + source + "\n"





r"""
Structural vector autoregression models
"""


#[

from __future__ import annotations

from .. import has_invariant as _has_invariant
from .. import has_variants as _has_variants

from ._invariants import Invariant
from ._variants import Variant

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Self
    from collections.abc import Iterable

#]


__all__ = (
    "StructVAR",
)


class StructVAR(
    _has_invariant.Mixin,
    _has_variants.Mixin,
):
    r"""
    """
    #[

    pass

    #]




"""
"""


__all__ = [
    "CURRENT_PORTABLE_FORMAT",
    "SUPPORTED_PORTABLE_FORMATS",
]


CURRENT_PORTABLE_FORMAT = "0.3.0"

SUPPORTED_LEGACY_PORTABLE_FORMATS = set()

SUPPORTED_PORTABLE_FORMATS = {CURRENT_PORTABLE_FORMAT, } | SUPPORTED_LEGACY_PORTABLE_FORMATS


def validate_portable_format(portable_format: str, /, ) -> None:
    if portable_format in SUPPORTED_PORTABLE_FORMATS:
        return
    raise _wrongdoings.IrisPieCritical(f"Portable format {portable_format} not supported")




"""
"""


#[
from __future__ import annotations

from typing import (Callable, )
import numpy as _np
import functools as _ft

from ..frames import (Frame, )
from ..dataslates.main import (Dataslate, )
from .solutions import (Solution, )
#]


def _simulate_anticipated_shock_values(
    model_v,
    dataslate_v: Dataslate,
    frame: Frame,
    get_solution_expansion: Callable,
) -> list[_np.ndarray | None]:
    """
    """
    #[
    #
    # Preallocate impact with None, and return immediately if no anticipated
    # shocks are present
    impact = [None] * dataslate_v.num_periods
    vec = model_v._get_dynamic_solution_vectors()
    if not vec.anticipated_shock_values:
        return impact
    #
    # Deviation is irrelevant for the impact of anticipated shocks
    solution = model_v._gets_solution()
    base_columns = dataslate_v.base_columns
    first_column = frame.start - dataslate_v.periods[0]
    last_column = frame.simulation_end - dataslate_v.periods[0]
    columns_to_run = tuple(range(first_column, last_column+1))
    working_data = dataslate_v.get_data_variant(0, )
    #
    v_array = extract_shock_values(
        working_data,
        vec.anticipated_shock_values,
    )
    #
    column_incidence = _np.any(v_array[:, columns_to_run], axis=0, ).tolist()
    if any(column_incidence, ):
        #
        # Last shock is at t+forward
        last_index_reversed = column_incidence[::-1].index(True)
        forward = len(column_incidence) - last_index_reversed - 1
        Rx = get_solution_expansion(solution, forward, )
        #
        # Last column within dataslate_v
        last_ant_column = first_column + forward
        #
        for t in columns_to_run:
            impact[t] = sum(
                Rx[k] @ v_array[:, s]
                for k, s in enumerate(range(t, last_ant_column+1))
            )
    return impact
    #]


simulate_square_anticipated_shock_values = _ft.partial(
    _simulate_anticipated_shock_values,
    get_solution_expansion=Solution.expand_square_solution,
)


simulate_triangular_anticipated_shock_values = _ft.partial(
    _simulate_anticipated_shock_values,
    get_solution_expansion=Solution.expand_triangular_solution,
)


def extract_shock_values(working_data, shock_vector, ) -> _np.ndarray:
    """
    """
    #[
    shock_qids = [t.qid for t in shock_vector]
    return working_data[shock_qids, :]
    #]




r"""
Implement SlatableProtocol for Stacker objects
"""


#[

from __future__ import annotations

from ..dataslates import Slatable
from ..series import Series
from .. import quantities as _quantities

#]


_DEFAULT_SHOCK_VALUE = 0.0


class Inlay:
    r"""
    """
    #[

    def slatable_for_marginal(
        self,
        shocks_from_data: bool,
        stds_from_data: bool,
    ) -> Slatable:
        r"""
        """
        #
        if self.num_variants != 1:
            raise ValueError("Not implemented for multiple variants")
        #
        slatable = Slatable()
        #
        slatable.max_lag = 0 # self.max_lag
        slatable.max_lead = 0 # self.max_lead
        #
        qid_to_name = self.create_qid_to_name()
        slatable.databox_names = tuple(
            qid_to_name[qid]
            for qid in sorted(qid_to_name)
        )
        #
        name_to_description = self.create_name_to_description()
        slatable.descriptions = tuple(
            name_to_description.get(name, "", )
            for name in slatable.databox_names
        )
        #
        # Databox validation - all variables must be time series
        variable_names = self.get_names(kind=_quantities.ANY_VARIABLE, )
        validator = (
            lambda x: isinstance(x, Series),
            "Input data for this variable is not a time series",
        )
        slatable.databox_validators = {
            name: validator
            for name in variable_names
        }
        #
        # Fallbacks and overwrites
        slatable.fallbacks = {}
        slatable.overwrites = {}
        #
        shock_names = self.get_names(kind=_quantities.ANY_SHOCK_OR_SHOCK_VALUE, )
        shock_name_to_value = {
            name: [_DEFAULT_SHOCK_VALUE, ]*self.num_variants
            for name in shock_names
        }
        if shocks_from_data:
            slatable.fallbacks.update(shock_name_to_value, )
        else:
            slatable.overwrites.update(shock_name_to_value, )
        #
        stds = self._variants[0].std_name_to_value
        if stds_from_data:
            slatable.fallbacks.update(stds, )
        else:
            slatable.overwrites.update(stds, )
        #
        slatable.qid_to_logly = self.create_qid_to_logly()
        #
        return slatable

    #]




r"""
# First-order solution matrices


## Square solution:

$$
\begin{gathered}
\xi_t = T \, \xi_{t-1} + P \, u_t + \sum \, R \, v_t + K
\\
y_t = Z \, \xi_t + H \, w_t + D
\end{gathered}
$$


## Equivalent block-triangular solution:

$$
\begin{gathered}
\alpha_t = T_\alpha \, \alpha_{t-1} + P_\alpha \, u_t + \sum \, R_\alpha \, v_t + K_\alpha
\\
y_t = Z_\alpha \, \alpha_t + D + H \, w_t
\\
\xi_t \equiv = U_alpha \, \alpha_t
\end{gathered}
$$


## Forward expansion:

$$
\cdots
$$
"""


#[

from __future__ import annotations

import warnings as _wa
import enum as _en
import numpy as _np
import scipy as _sp
import copy as _co

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Self, Callable, Iterable
    from numbers import Real
    from ..fords import descriptors as _descriptors
    from ..fords import systems as _systems

#]


__all__ = (
    "STABLE", "UNIT_ROOT", "UNSTABLE", "UNIT",
)


class UnitRootException(Exception):
    pass


class EigenvalueKind(_en.Flag, ):
    STABLE = _en.auto()
    UNIT_ROOT = _en.auto()
    UNSTABLE = _en.auto()
    UNIT = UNIT_ROOT
    ALL = STABLE | UNIT_ROOT | UNSTABLE


class SystemStabilityKind(_en.Flag, ):
    STABLE = _en.auto()
    MULTIPLE_STABLE = _en.auto()
    NO_STABLE = _en.auto()


class Solution:
    """
    ## Square solution:

    T: Transition matrix
    P: Impact matrix of transition shocks
    K: Intercept in transition equation
    Z: Measurement matrix
    H: Impact matrix of measurement shocks
    D: Intercept in measurement equation


    ## Equivalent block-triangular solution:

    Ta: Transition matrix in triangular system
    Pa: Impact matrix of transition shocks in triangular system
    Ka: Intercept in transition equation in triangular system
    Za: Measurement matrix in triangular system
    Ua: Rotation matrix from triangular to square system


    ## Forward expansion of square solution:

    J: Power matrix
    Ru: Forward-looking impact matrix of transition shocks
    X: Impact matrix in square system
    Xa: Impact matrix in triangular system
    """
    #[

    __slots__ = (
        "T", "P", "K", "Z", "H", "D",
        "Ta", "Pa", "Ka", "Za", "Ua",
        "J", "Ru", "X", "Xa",
        "square_expansion",
        "triangular_expansion",
        "eigenvalues",

        "eigenvalues_stability",
        "system_stability",
        "transition_vector_stability",
        "measurement_vector_stability",
    )

    def __init__(self, ) -> None:
        r"""
        """
        for n in self.__slots__:
            setattr(self, n, None, )

    @classmethod
    def from_system(
        klass,
        descriptor: _descriptors.Descriptor,
        system: _systems.System,
        tolerance: float,
        clip_small: bool,
    ) -> Self:
        """
        """
        self = klass()
        #
        def is_alpha_beta_stable_or_unit_root(alpha: Real, beta: Real, ) -> bool:
            abs_alpha = abs(alpha)
            abs_beta = abs(beta)
            return abs_beta < (1 + tolerance)*abs_alpha
        #
        def is_stable_root(root: complex, ) -> bool:
            abs_root = abs(root)
            return abs_root < (1 - tolerance)
        #
        def is_unit_root(root: complex, ) -> bool:
            abs_root = abs(root)
            return abs_root >= (1 - tolerance) and abs_root < (1 + tolerance)
        #
        def clip_func(x: _np.ndarray, ) -> _np.ndarray:
            return _np.where(_np.abs(x) < tolerance, 0, x)
        #
        clip = clip_func if clip_small else None
        #
        # Detach unstable from (stable + unit) roots and solve out expectations
        # The system is triangular but because stable and unit roots are
        # not detached yet, the system is called "preliminary"
        qz_matrixes, self.eigenvalues = _solve_ordqz(
            system,
            is_alpha_beta_stable_or_unit_root,
        )
        #
        # Classify eigenvalues as stable, unit root, or unstable
        self._classify_eigenvalues_stability(is_stable_root, is_unit_root, )
        #
        triangular_solution_prelim = \
            _solve_transition_equations(descriptor, system, qz_matrixes, )
        #
        # Detach unit from stable roots to create the final triangular solution
        # From the final triangular solution, calculate the square solution
        *triangular_solution, check_num_unit_roots = detach_stable_from_unit_roots(
            triangular_solution_prelim,
            is_unit_root,
            clip=clip,
        )
        if check_num_unit_roots != self.num_unit_roots:
            raise UnitRootException
        #
        square_solution = _square_from_triangular(triangular_solution, )
        #
        self.Ua, self.Ta, self.Pa, self.Ka, self.Xa, self.J, self.Ru = triangular_solution
        self.T, self.P, self.K, self.X = square_solution
        #
        # Solve measurement equations
        self.Z, self.H, self.D, self.Za = _solve_measurement_equations(
            descriptor,
            system,
            self.Ua,
            clip=clip,
        )
        self._classify_system_stability(descriptor.get_num_forwards(), )
        self._classify_transition_vector_stability(tolerance=tolerance, )
        self._classify_measurement_vector_stability(tolerance=tolerance, )
        #
        self.square_expansion = []
        self.triangular_expansion = []
        #
        return self

    def create_deviation_solution(self, ) -> Self:
        r"""
        Create a shallow copy of the solution, and replace constant vectors with
        zeros
        """
        new = type(self)()
        for n in new.__slots__:
            setattr(new, n, getattr(self, n, None), )
        if new.K is not None:
            new.K = _np.zeros_like(self.K, )
        if new.Ka is not None:
            new.Ka = _np.zeros_like(self.Ka, )
        if new.D is not None:
            new.D = _np.zeros_like(self.D, )
        return new

    @property
    def num_xi(self, ) -> int:
        """==Number of xi vector elements=="""
        return self.T.shape[0]

    @property
    def num_alpha(self, ) -> int:
        """==Number of alpha vector elements=="""
        return self.Ta.shape[0]

    @property
    def num_y(self, ) -> int:
        """==Number of y vector elements=="""
        return self.Z.shape[0]

    @property
    def num_u(self, ) -> int:
        """==Number of u vector elements=="""
        return self.P.shape[1]

    @property
    def num_v(self, ) -> int:
        """==Number of v vector elements=="""
        return self.P.shape[1]

    @property
    def num_w(self, ) -> int:
        """==Number of w vector elements=="""
        return self.H.shape[1]

    @property
    def num_unit_roots(self, ) -> int:
        """==Number of unit roots=="""
        return self.eigenvalues_stability.count(EigenvalueKind.UNIT_ROOT)

    @property
    def num_stable(self, ) -> int:
        """==Number of stable elements in alpha vector=="""
        return self.num_alpha - self.num_unit_roots

    @property
    def Ta_stable(self, ) -> _np.ndarray:
        """==Stable part of transition matrix=="""
        num_unit_roots = self.num_unit_roots
        return self.Ta[num_unit_roots:, num_unit_roots:]

    @property
    def Pa_stable(self, ) -> _np.ndarray:
        """==Stable part of impact matrix of transition shocks=="""
        num_unit_roots = self.num_unit_roots
        return self.Pa[num_unit_roots:, :]

    @property
    def Ka_stable(self, ) -> _np.ndarray:
        """==Stable part of intercept in transition equation=="""
        num_unit_roots = self.num_unit_roots
        return self.Ka[num_unit_roots:]

    @property
    def Za_stable(self, ) -> _np.ndarray:
        """==Stable part of measurement matrix=="""
        num_unit_roots = self.num_unit_roots
        return self.Za[:, num_unit_roots:]

    @property
    def boolex_stable_transition_vector(self, ) -> tuple[int, ...]:
        r"""==Index of stable transition vector elements=="""
        return _np.array(tuple(
            i == EigenvalueKind.STABLE
            for i in self.transition_vector_stability
        ), dtype=bool, )

    @property
    def boolex_stable_measurement_vector(self, ) -> tuple[int, ...]:
        r"""==Index of stable measurement vector elements=="""
        return _np.array(tuple(
            i == EigenvalueKind.STABLE
            for i in self.measurement_vector_stability
        ), dtype=bool, )

    def unpack_square_solution(self, ) -> tuple[_np.ndarray, ...]:
        r"""
        Return square solution matrices in the following order:
        T, P, K, Z, H, D, None
        """
        return self.T, self.P, self.K, self.Z, self.H, self.D, None,

    def unpack_triangular_solution(self, ) -> tuple[_np.ndarray, ...]:
        r"""
        Return triangular solution matrices in the following order:
        Ta, Pa, Ka, Za, H, D, Ua
        """
        return self.Ta, self.Pa, self.Ka, self.Za, self.H, self.D, self.Ua,

    def copy(self, ) -> Self:
        r"""
        """
        return _co.deepcopy(self, )

    def expand_square_solution(self, forward: int, ) -> list[_np.ndarray]:
        r"""
        Expand R matrices of square solution for t+1...t+forward
        """
        return _get_solution_expansion(
            self.square_expansion,
            self.P, self.X, self.J, self.Ru,
            forward,
        )

    def expand_triangular_solution(self, forward: int, ) -> list[_np.ndarray]:
        """
        Expand Ra matrices of square solution for t+1...t+forward
        """
        return _get_solution_expansion(
            self.triangular_expansion,
            self.Pa, self.Xa, self.J, self.Ru,
            forward,
        )

    def _classify_eigenvalues_stability(
        self,
        is_stable_root: Callable[[complex], bool],
        is_unit_root: Callable[[complex], bool],
    ) -> None:
        self.eigenvalues_stability = tuple(
            _classify_eigenvalue_stability(v, is_stable_root, is_unit_root, )
            for v in self.eigenvalues
        )

    def _classify_system_stability(
        self,
        num_forwards: int,
    ) -> None:
        num_unstable = self.eigenvalues_stability.count(EigenvalueKind.UNSTABLE)
        if num_unstable == num_forwards:
            self.system_stability = SystemStabilityKind.STABLE
        elif num_unstable > num_forwards:
            self.system_stability = SystemStabilityKind.NO_STABLE
        else:
            self.system_stability = SystemStabilityKind.MULTIPLE_STABLE

    def _classify_transition_vector_stability(
        self,
        tolerance: float,
    ) -> None:
        self.transition_vector_stability \
            = _classify_solution_vector_stability(
                self.Ua,
                self.num_unit_roots,
                tolerance=tolerance,
            )

    def _classify_measurement_vector_stability(
        self,
        tolerance: float,
    ) -> None:
        self.measurement_vector_stability \
            = _classify_solution_vector_stability(
                self.Za,
                self.num_unit_roots,
                tolerance=tolerance,
            )
    #]


def left_div(A: _np.ndarray, B: _np.ndarray, ) -> _np.ndarray:
    r"""
    Solve A \ B = pinv(A) @ B or inv(A) @ B
    """
    return _np.linalg.lstsq(A, B, rcond=None)[0]


def right_div(B: _np.ndarray, A: _np.ndarray, ) -> _np.ndarray:
    r"""
    Solve B / A which is (A' \ B')'
    """
    return _np.linalg.lstsq(A.T, B.T, rcond=None)[0].T


def _square_from_triangular(
    triangular_solution: tuple[_np.ndarray, ...],
) -> tuple[_np.ndarray, ...]:
    r"""
    T <- Ua @ Ta / Ua
    R <- Ua @ Ra
    X <- Xa @ Ra
    K <- Ua @ Ka
    xi[t] = ... -X J**(k-1) Ru e[t+k]
    """
    #[
    Ua, Ta, Ra, Ka, Xa, *_ = triangular_solution
    T = Ua @ right_div(Ta, Ua) # Ua @ (Ta / Ua)
    R = Ua @ Ra
    K = Ua @ Ka
    X = Ua @ Xa
    return T, R, K, X,
    #]


def detach_stable_from_unit_roots(
    transition_solution_prelim: tuple[_np.ndarray, ...],
    is_unit_root: Callable[[Real], bool],
    clip: Callable | None,
) -> tuple[_np.ndarray, ...]:
    """
    Apply a secondary Schur decomposition to detach stable eigenvalues from unit roots
    """
    #[
    Ug, Tg, Rg, Kg, Xg, J, Ru = transition_solution_prelim
    num_xib = Tg.shape[0]
    Ta, u, check_num_unit_roots = _sp.linalg.schur(Tg, sort=is_unit_root, ) # Tg = u @ Ta @ u.T
    Ua = Ug @ u if Ug is not None else u
    Ua = clip(Ua) if clip is not None else Ua
    Ra = u.T @ Rg
    Ka = u.T @ Kg
    Xa = u.T @ Xg if Xg is not None else None
    return Ua, Ta, Ra, Ka, Xa, J, Ru, check_num_unit_roots,
    #]


def _solve_measurement_equations(
    descriptor,
    system,
    Ua,
    *,
    clip: Callable | None,
) -> tuple[_np.ndarray, ...]:
    r"""
    """
    #[
    num_forwards = descriptor.get_num_forwards()
    G = system.G[:, num_forwards:]
    Z = left_div(-system.F, G) # -F \ G
    H = left_div(-system.F, system.J) # -F \ J
    D = left_div(-system.F, system.H) # -F \ H
    Z = clip(Z) if clip is not None else Z
    Za = Z @ Ua
    return Z, H, D, Za,
    #]


def _solve_transition_equations(
    descriptor,
    system,
    qz_matrixes: tuple[_np.ndarray, ...],
) -> tuple[_np.ndarray, ...]:
    r"""
    """
    #[
    num_backwards = descriptor.get_num_backwards()
    num_forwards = descriptor.get_num_forwards()
    num_stable = num_backwards
    S, T, Q, Z, = qz_matrixes
    #
    S11 = S[:num_stable, :num_stable]
    S12 = S[:num_stable, num_stable:]
    S22 = S[num_stable:, num_stable:]
    #
    T11 = T[:num_stable, :num_stable]
    T12 = T[:num_stable, num_stable:]
    T22 = T[num_stable:, num_stable:]
    #
    Z21 = Z[num_forwards:, :num_stable]
    Z22 = Z[num_forwards:, num_stable:]
    #
    # Constant in transition equations
    Q_CC = Q @ system.C
    Q_CC1 = Q_CC[:num_stable]
    Q_CC2 = Q_CC[num_stable:]
    #
    # Transition shocks in transition equations
    Q_DD = Q @ system.D
    Q_DD1 = Q_DD[:num_stable, :]
    Q_DD2 = Q_DD[num_stable:, :]
    #
    # Unstable block
    #
    G = left_div(-Z21, Z22) # -Z21 \ Z22
    Ru = left_div(-T22, Q_DD2) # -T22 \ Q_DD2
    Ku = left_div(-(S22 + T22), Q_CC2) # -(S22+T22) \ Q_CC2
    #
    # Transform stable block==transform backward-looking variables:
    # gamma(t) = s(t) + G u(t+1)
    #
    Xg0 = left_div(S11, T11 @ G + T12)
    Xg1 = G + left_div(S11, S12)
    #
    Tg = left_div(-S11, T11)
    Rg = -Xg0 @ Ru - left_div(S11, Q_DD1)
    Kg = -(Xg0 + Xg1) @ Ku - left_div(S11, Q_CC1)
    Ug = Z21 # xib = Ug @ gamma
    #
    # Forward expansion
    # gamma(t) = ... -Xg J**(k-1) Ru e(t+k)
    #
    J = left_div(-T22, S22) # -T22 \ S22
    Xg = Xg1 + Xg0 @ J
    #
    return Ug, Tg, Rg, Kg, Xg, J, Ru,
    #]


def _solve_ordqz(
    system: _systems.System,
    is_alpha_beta_stable_or_unit_root: Callable,
) -> tuple[tuple[_np.ndarray, ...], tuple[complex, ...], ]:
    r"""
    Solve the system using the ordered QZ decomposition and separate unstable
    roots from stable and unit roots
    """
    #[
    #
    # Calculate QZ decomposition
    S, T, alpha, beta, Q, Z = _sp.linalg.ordqz(
        system.A, system.B,
        sort=is_alpha_beta_stable_or_unit_root,
    )
    Q = Q.T
    qz_matrixes = (S, T, Q, Z, )
    #
    # Calculate eigenvalues
    _wa.filterwarnings(action="ignore", category=RuntimeWarning, )
    eigenvalues = tuple(complex(i) for i in -beta/alpha)
    _wa.filterwarnings(action="default", category=RuntimeWarning, )
    #
    return qz_matrixes, eigenvalues,
    #]


def _classify_eigenvalue_stability(
    eigenvalue: complex,
    is_stable_root: Callable[[complex], bool],
    is_unit_root: Callable[[complex], bool],
) -> EigenvalueKind:
    r"""
    Classify a complex number (eigenvalue) as stable, unit root, or unstable.
    """
    #[
    abs_eigenvalue = _np.abs(eigenvalue)
    if is_stable_root(abs_eigenvalue):
        return EigenvalueKind.STABLE
    elif is_unit_root(abs_eigenvalue):
        return EigenvalueKind.UNIT_ROOT
    else:
        return EigenvalueKind.UNSTABLE
    #]


def _classify_solution_vector_stability(
    transform_matrix: _np.ndarray,
    num_unit_roots: int,
    tolerance: float,
) -> None:
    """
    """
    #[
    test_matrix = _np.abs(transform_matrix[:, :num_unit_roots], )
    index = _np.any(test_matrix > tolerance, axis=1, )
    return tuple(
        EigenvalueKind.UNIT_ROOT if i
        else EigenvalueKind.STABLE
        for i in index
    )
    #]


def _get_solution_expansion(
    existing_expansion: list[_np.ndarray],
    P, X, J, Ru,
    forward: int,
) -> list[_np.ndarray]:
    """
    Expand R matrices of square solution for t+1...t+forward
    """
    if (P is None) or (X is None) or (J is None) or (Ru is None):
        return None
    #
    # return [R(t), R(t+1), R(t+2), ..., R(t+forward)]
    #
    # R(t) = R
    # R(t+k) = -X J**(k-1) Ru e(t+k)
    # k = 1, ..., forward or k-1 = 0, ..., forward-1
    #
    R0 = _np.array(P, )
    existing_forward = len(existing_expansion)
    for k_minus_1 in range(existing_forward, forward):
        Rk = -X @ _np.linalg.matrix_power(J, k_minus_1, ) @ Ru
        existing_expansion.append(Rk, )
    return [R0, ] + existing_expansion[:forward]
    #
    # return [R, ] + [
    #     -X @ _np.linalg.matrix_power(J, k_minus_1) @ Ru
    #     for k_minus_1 in range(0, forward)
    # ]


STABLE = EigenvalueKind.STABLE
UNIT_ROOT = EigenvalueKind.UNIT_ROOT
UNSTABLE = EigenvalueKind.UNSTABLE
UNIT = UNIT_ROOT




"""
"""


#[

from __future__ import annotations

from .. import dates as _dates
from ..dates import Period
from ..fords.descriptors import SolutionVectors, Squid
from ..simultaneous.main import Simultaneous
from .. import quantities as _quantities
from ..quantities import QuantityKind, Quantity
from ..incidences.main import Token

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence
    from typing import Self, Any

#]


class Invariant:
    r"""
    """
    #[

    __slots__ = (
        "base_periods",
        "transition_period_indexes",
        "measurement_period_indexes",
        "quantities",
        "max_lag",
        "max_lead",
        "source_solution_vectors",
        "stacked_solution_vectors",
        "squid",
        "index_xi",
        "index_u",
        "index_v",
        "index_y",
        "index_w",
    )

    def __init__(self, **kwargs, ) -> None:
        """
        """
        for slot in self.__slots__:
            setattr(self, slot, kwargs.get(slot, None, ), )

    @classmethod
    def from_simultaneous(
        klass,
        model: Simultaneous,
        span: Iterable[Period],
        *,
        transition_variables: None | Iterable[str] = None,
        anticipated_shocks: None | Iterable[str] = None,
        unanticipated_shocks: None | Iterable[str] = None,
        measurement_variables: None | Iterable[str] = None,
        measurement_shocks: None | Iterable[str] = None,
    ) -> Self:
        """
        """
        self = klass()
        self._populate_periods(span, )
        self.quantities = model.quantities
        self.max_lag = model.max_lag
        self.max_lead = model.max_lead
        self.source_solution_vectors = model._invariant.dynamic_descriptor.solution_vectors
        #
        name_to_qid = model.create_name_to_qid()
        quantity_names = {
            "transition_variables": transition_variables,
            "anticipated_shocks": anticipated_shocks,
            "unanticipated_shocks": unanticipated_shocks,
            "measurement_variables": measurement_variables,
            "measurement_shocks": measurement_shocks,
        }
        for kind in quantity_names.keys():
            if quantity_names[kind] is None:
                quantity_names[kind] = model.get_names(kind=QuantityKind.from_keyword(kind, ), )
        #
        quantity_ids = { kind: None for kind in quantity_names.keys() }
        self.stacked_solution_vectors = SolutionVectors()
        for kind in quantity_names.keys():
            sorted_qids = sorted(set(tuple(name_to_qid[name] for name in quantity_names[kind])))
            vector = tuple( Token(qid, 0, ) for qid in sorted_qids )
            setattr(self.stacked_solution_vectors, kind, vector, )
        #
        self._populate_indexes()
        self._populate_squid(model, )
        #
        return self

    @property
    def num_periods(self, ) -> int:
        r"""
        """
        return len(self.base_periods)

    def _populate_periods(self, span: Iterable[Period], ) -> None:
        r"""
        """
        span = tuple(i for i in span)
        self.base_periods = _dates.periods_from_until(span[0], span[-1], )

    def _populate_indexes(self, ) -> None:
        self.index_xi = _create_index(
            self.source_solution_vectors.transition_variables,
            self.stacked_solution_vectors.transition_variables,
        )
        self.index_u = _create_index(
            self.source_solution_vectors.unanticipated_shocks,
            self.stacked_solution_vectors.unanticipated_shocks,
        )
        self.index_v = _create_index(
            self.source_solution_vectors.anticipated_shocks,
            self.stacked_solution_vectors.anticipated_shocks,
        )
        self.index_y = _create_index(
            self.source_solution_vectors.measurement_variables,
            self.stacked_solution_vectors.measurement_variables,
        )
        self.index_w = _create_index(
            self.source_solution_vectors.measurement_shocks,
            self.stacked_solution_vectors.measurement_shocks,
        )

    def _populate_squid(self, model, ) -> None:
        r"""
        """
        squid = Squid.from_squidable(model, )
        squid.y_qids = _extract_by_index(squid.y_qids, self.index_y, )
        squid.u_qids = _extract_by_index(squid.u_qids, self.index_u, )
        squid.v_qids = _extract_by_index(squid.v_qids, self.index_v, )
        squid.w_qids = _extract_by_index(squid.w_qids, self.index_w, )
        squid.std_u_qids = _extract_by_index(squid.std_u_qids, self.index_u, )
        squid.std_v_qids = _extract_by_index(squid.std_v_qids, self.index_v, )
        squid.std_w_qids = _extract_by_index(squid.std_w_qids, self.index_w, )
        self.squid = squid

    #]


def _create_index(source_vector, stacked_vector, ) -> list[int, ...]:
    r"""
    Return index of positions in source_vector for each token in stacked_vector.
    Needs to be a list, not a tuple, because it is used for indexing in numpy.
    """
    return [
        source_vector.index(tok)
        for tok in stacked_vector
    ]


def _extract_by_index(
    source: Sequence[Any],
    index: list[int],
) -> list[Any, ...]:
    r"""
    Return a list of elements from a source sequence at positions given by index
    """
    return [ source[i] for i in index ]




r"""
"""


#[

from __future__ import annotations

import functools as _ft
import numpy as _np

from ..databoxes import Databox
from ..dataslates import Dataslate
from ..dates import Period
from ..fords import least_squares as _least_squares
from .. import dates as _periods

from ._variants import Variant
from . import prior_obs as _prior_obs
from .prior_obs import PriorObs
from ._dimensions import Dimensions
from ..progress_bars import ProgressBar
from ..fords import covariances as _covariances

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Iterable, Callable, Literal

#]


_GET_SPANS_DISPATCH = {
    "short": _periods.spans_from_short_span,
    "long": _periods.spans_from_long_span,
}


class Inlay:
    r"""
    """
    #[

    def estimate(
        self,
        input_data: Databox,
        span: Iterable[Period] | None = None,
        #
        interpret_span: Literal["short", "long"] = "short",
        num_variants: int | None = None,
        omit_missing: bool = True,
        prior_obs: PriorObs | Iterable[PriorObs, ] | None = None,
        dof_correction: bool = False,
        target_db: Databox | None = None,
        show_progress: bool = False,
        progress_bar_settings: dict = dict(title="Estimating RedVAR", ),
    ) -> Databox:
        r"""
        """
        dimensions = self._invariant.dimensions
        has_intercept = dimensions.has_intercept
        short_span, long_span, = _GET_SPANS_DISPATCH[interpret_span](span, )
        num_variants = self.resolve_num_variants_in_context(num_variants, )
        slatable = self.slatable_for_estimate()
        dataslate = Dataslate.from_databox_for_slatable(
            slatable, input_data, short_span,
            num_variants=num_variants,
        )
        #
        zipped = zip(
            range(num_variants, ),
            dataslate.iter_variants(),
        )
        self._variants = []
        #
        # Estimate each variant
        #=======================================================================
        progress_bar = ProgressBar(
            num_steps=num_variants,
            show_progress=show_progress,
            **progress_bar_settings,
        )
        for vid, dataslate_v in zipped:
            estimated_variant = _estimate_variant(
                self._invariant,
                dataslate_v,
                omit_missing=omit_missing,
                prior_obs=prior_obs,
                dof_correction=dof_correction,
            )
            self._variants.append(estimated_variant, )
            progress_bar.increment()
        #=======================================================================
        #
        output_db = dataslate.to_databox()
        if target_db is not None:
            output_db = target_db | output_db
        return output_db
    #]


def _estimate_variant(
    invariant: Invariant,
    dataslate: Dataslate,
    prior_obs: PriorObs | Iterable[PriorObs] | None,
    dof_correction: bool,
    omit_missing: bool,
) -> Variant:
    r"""
    """
    #[
    dimensions = invariant.dimensions
    order = dimensions.order
    has_intercept = dimensions.has_intercept
    num_endogenous = dimensions.num_endogenous
    num_rhs = dimensions.num_nonendogenous
    num_lagged_endogenous = dimensions.num_lagged_endogenous
    #
    y0, y1, x, k, where, = _get_estimation_data(invariant, dataslate, omit_missing, )
    num_periods_fitted = int(where.sum())
    if num_periods_fitted == 0:
        raise ValueError("No data available for estimation after removing periods with missing observations.")
    #
    lhs_est = y0[:, where]
    rhs_est = _np.vstack([y1, x, k, ])[:, where]
    #
    if prior_obs is not None:
        xk = _np.vstack([x, k, ], )
        gamma = _least_squares.ordinary_least_squares(y0[:, where], xk[:, where], )
        y0_demeaned = y0[:, where] - gamma @ xk[:, where]
        y0_std = _np.sqrt(_np.diag(y0_demeaned @ y0_demeaned.T / num_periods_fitted))
        y0_std = y0_std.reshape((-1, ), )
        lhs_dummy, rhs_dummy, = _prior_obs.arrays_from_prior_obs(prior_obs, dimensions, y0_std, )
        lhs_est = _np.hstack([lhs_est, lhs_dummy, ])
        rhs_est = _np.hstack([rhs_est, rhs_dummy, ])
    #
    beta = _least_squares.ordinary_least_squares(lhs_est, rhs_est, )
    #
    A = beta[:, :num_lagged_endogenous]
    if has_intercept:
        B = beta[:, num_lagged_endogenous:-1]
        c = beta[:, -1]
    else:
        B = beta[:, num_lagged_endogenous:]
        c = None
    #
    u = y0 - A @ y1 - B @ x - c.reshape((-1, 1))
    #
    num_periods_corrected = num_periods_fitted - (num_rhs if dof_correction else 0)
    cov_residuals = u[:, where] @ u[:, where].T / num_periods_corrected
    cov_residuals = _covariances.symmetrize(cov_residuals, )
    #
    fitted_periods = tuple(
        period for period, been_fitted in zip(dataslate.base_periods, where, )
        if been_fitted
    )
    #
    estimated_variant = Variant(
        A=A,
        B=B,
        c=c,
        cov_residuals=cov_residuals,
        fitted_periods=fitted_periods,
        residual_estimates=u,
    )
    #
    _write_residual_estimates(
        invariant,
        dataslate,
        u,
    )
    #
    return estimated_variant
    #]


def _get_estimation_data(
    invariant: Invariant,
    dataslate: Dataslate,
    omit_missing: bool,
) -> tuple[_np.ndarray, _np.ndarray, _np.ndarray]:
    r"""
    """
    #[
    data = dataslate.get_data_variant()
    order = invariant.dimensions.order
    has_intercept = invariant.dimensions.has_intercept
    endogenous_qids = list(invariant.get_endogenous_qids())
    exogenous_qids = list(invariant.get_exogenous_qids())
    y = data[endogenous_qids, :]
    y0 = y[:, order:]
    y1 = _np.vstack([
        y[:, order-i:-i]
        for i in range(1, order+1, )
    ])
    x = data[exogenous_qids, order:]
    k = _np.ones((int(has_intercept), x.shape[1], ), dtype=float, )
    if omit_missing:
        where = get_where_observations([y0, y1, x, k], )
    else:
        num_periods = y0.shape[1]
        where = _np.ones(num_periods, dtype=bool, )
    #
    return y0, y1, x, k, where,
    #]


def _write_residual_estimates(
    invariant: Invariant,
    dataslate: Dataslate,
    residual_estimates: _np.ndarray,
) -> None:
    r"""
    """
    #[
    order = invariant.dimensions.order
    residual_qids = invariant.get_residual_qids()
    data_array = dataslate.get_data_variant()
    data_array[residual_qids, order:] = residual_estimates
    #]


def get_where_observations(data_matrices: list[_np.ndarray], ) -> _np.ndarray:
    r"""
    """
    return _np.all(_np.isfinite(_np.vstack(data_matrices, ), ), axis=0, )


def _add_intercept_dummy(x: _np.ndarray, ) -> _np.ndarray:
    r"""
    """
    return _np.pad(
        x,
        ((0, 1), (0, 0), ),
        mode="constant",
        constant_values=1,
    )




"""
Reduced-form vector autoregression (RedVAR) moduls
"""


from .main import *
from .main import __all__ as main_all

from .prior_obs import *
from .prior_obs import __all__ as prior_obs_all


__all__ = (
    *main_all,
    *prior_obs_all,
)




"""
Utility iterators
"""


#[
from __future__ import annotations

from collections.abc import (Iterable, Iterator, )
from typing import (Any, )
#]


def exhaust_then_last(
    iterable: Iterable[Any],
    default=None,
) -> Iterator[Any]:
    """
    Repeat the last element of an iterable indefinitely
    """
    last = default
    for item in iterable:
        yield item
        last = item
    while True:
        yield last




r"""
"""


#[

from __future__ import annotations

import numpy as np
import networkx as nx
from itertools import product

#]


def dulmage_mendelsohn_reordering(A: np.ndarray) -> tuple[list[int], list[int], np.ndarray]:
    """
    Reorders the incidence matrix using the Dulmage-Mendelsohn decomposition
    to produce the maximal number of smallest sequentially independent blocks.

    Parameters:
        A: np.ndarray
            Binary incidence matrix with shape (m equations, n variables)

    Returns:
        A tuple (row_order, col_order, A_reordered) where:
        - row_order: list of row indices
        - col_order: list of column indices
        - A_reordered: incidence matrix reordered according to row/column permutations
    """
    m, n = A.shape
    B = nx.Graph()

    # Construct bipartite graph: equations as ('e', i), variables as ('v', j)
    eq_idxs, var_idxs = np.where(A)
    for i, j in zip(eq_idxs, var_idxs):
        B.add_edge(("e", int(i)), ("v", int(j)))

    # Find maximum matching on the bipartite graph
    matching = nx.bipartite.maximum_matching(B, top_nodes={("e", i) for i in range(m)})

    # Build directed graph for Dulmage-Mendelsohn decomposition
    G = nx.DiGraph()
    for u, v in B.edges():
        # Reverse direction if edge is in the matching
        if matching.get(u) == v or matching.get(v) == u:
            if u[0] == "e":
                G.add_edge(v, u)
            else:
                G.add_edge(u, v)
        else:
            # Keep direction from equation to variable otherwise
            if u[0] == "e":
                G.add_edge(u, v)
            else:
                G.add_edge(v, u)

    # Unmatched equations and variables
    unmatched_eqs = [("e", i) for i in range(m) if ("e", i) not in matching]
    unmatched_vars = [("v", j) for j in range(n) if ("v", j) not in matching]

    # Compute forward reachable nodes from unmatched equations
    F_reachable = set()
    for e in unmatched_eqs:
        F_reachable.update(nx.descendants(G, e) | {e})

    # Compute backward reachable nodes from unmatched variables (in reversed graph)
    GT = G.reverse()
    T_reachable = set()
    for v in unmatched_vars:
        T_reachable.update(nx.descendants(GT, v) | {v})

    # Remaining nodes define well-determined block(s)
    remaining = set(G.nodes()) - F_reachable - T_reachable

    # Find strongly connected components in the well-determined region
    subgraph = G.subgraph(remaining).copy()
    sccs = list(nx.strongly_connected_components(subgraph))

    # Collect nodes by type from SCCs
    block_eqs = []
    block_vars = []
    block_list = []
    for component in sccs:
        eqs = sorted([n[1] for n in component if n[0] == "e"])
        vars_ = sorted([n[1] for n in component if n[0] == "v"])
        block_eqs.extend(eqs)
        block_vars.extend(vars_)
        block_list.append((eqs, vars_))

    # Extract nodes from F and T partitions
    F_eqs = [n[1] for n in F_reachable if n[0] == "e"]
    F_vars = [n[1] for n in F_reachable if n[0] == "v"]
    T_eqs = [n[1] for n in T_reachable if n[0] == "e"]
    T_vars = [n[1] for n in T_reachable if n[0] == "v"]

    # Final row/column permutation orders
    row_order = F_eqs + block_eqs + T_eqs
    col_order = F_vars + block_vars + T_vars

    # Apply permutation to matrix
    A_reordered = A[np.ix_(row_order, col_order)]

    return row_order, col_order, A_reordered

def extract_dm_blocks(A: np.ndarray, row_order: list[int], col_order: list[int]) -> list[dict[str, object]]:
    """
    Splits a reordered matrix into sequential blocks assuming DM reordering.

    Parameters:
        A: np.ndarray
            Original incidence matrix
        row_order: list[int]
            Row permutation from DM reordering
        col_order: list[int]
            Column permutation from DM reordering

    Returns:
        A list of dictionaries with keys:
        - 'block': submatrix for the block
        - 'rows': list of original row indices in the block
        - 'cols': list of original column indices in the block
    """
    A_reordered = A[np.ix_(row_order, col_order)]

    blocks = []
    start_row = 0
    for i in range(1, len(row_order) + 1):
        if i == len(row_order) or not np.any(A_reordered[start_row:i, :]):
            if i > start_row:
                block = A_reordered[start_row:i, :]
                if block.any():
                    blocks.append({
                        'block': block,
                        'rows': row_order[start_row:i],
                        'cols': col_order
                    })
            start_row = i

    return blocks

# Example usage:
if __name__ == '__main__':
    A = np.array([
        [1, 1, 0, 0],
        [0, 1, 1, 0],
        [0, 0, 1, 1],
        [1, 0, 0, 1]
    ], dtype=bool)
    row_order, col_order, A_reordered = dulmage_mendelsohn_reordering(A)
    print("Row order:", row_order)
    print("Column order:", col_order)
    print("Reordered matrix:\n", A_reordered)

    blocks = extract_dm_blocks(A, row_order, col_order)
    for idx, blk in enumerate(blocks):
        print(f"\nBlock {idx} shape: {blk['block'].shape}")
        print("Rows:", blk['rows'])
        print("Cols:", blk['cols'])
        print(blk['block'])




"""
Merge inlay
"""


#[

from __future__ import annotations

from typing import Literal
import warnings as _wa
import documark as _dm

from .. import wrongdoings as _wrongdoings
from ..series import main as _series
from . import main as _databoxes

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Self, Iterable

#]


MergeStrategyType = Literal[
    "stack", # Stack values as variants
    "hstack", # Legacy alias for stack, do not include in the docstring
    "replace", # Replace the existing value with the new one
    "discard", # Discard the new value and keep the existing one
    "silent", # Exactly the same as discard
    "warning", # Same as discard with a warning
    "error", # Throw an error for the first duplicate key
    "critical", # Throw a critical error for the first duplicate key
]


def mixin(klass, ):
    r"""
    Mix the by_merging and merge methods into the class
    """
    #[
    klass.by_merging = classmethod(_by_merging, )
    klass.merge = _merge
    return klass
    #]


def _by_merging(
    klass,
    databoxes: Iterable[Self],
    merge_strategy: MergeStrategyType = "stack",
) -> Self:
    r"""
    """
    self = klass()
    self.merge(databoxes, merge_strategy, )
    return self


@_dm.reference(category="multiple", )
def _merge(
    self: Self,
    other: Self | Iterable[Self],
    merge_strategy: MergeStrategyType = "stack",
    # Do not include the following in the docstring
    action = None,
) -> None:
    r"""
................................................................................

==Merge Databoxes==

Combine one or more databoxes into a single databox using a specified merge
strategy to handle potential conflicts between duplicate keys.

self.merge(
    other,
    merge_strategy="stack",
)


### Input arguments ###

???+ input "other"
The databox or iterable of databoxes to merge into the current databox. If
merging a single databox, it should be passed directly; for multiple
databoxes, pass an iterable containing all.

???+ input "merge_strategy"
Determines how to process keys that exist in more than one databox. The
default strategy is `"stack"`.

* `"stack"`: Stack values; this means combine time series into multiple
columns, or combine lists, or convert non-lists to lists for stacking.

* `"replace"`: Replace existing values with new values.

* `"discard"` and `"silent"`: Retain original values and ignore new values.

* `"warning"`: Behave like `"discard"` but issue a warning for each conflict.

* `"error"`: Raise an error on encountering the first duplicate key.

* `"critical"`: Raise a critical error on encountering the first duplicate key.


### Returns ###

This method modifies the databox in place and returns `None`.

................................................................................
    """
    # Legacy name
    if action is not None:
        _wa.warn("The 'action' input argument is deprecated; use 'merge_strategy' instead", )
        merge_strategy = action
    #
    merge_strategy_func = _MERGE_STRATEGY_DISPATCH[merge_strategy]
    stream = _wrongdoings.create_stream(
        merge_strategy,
        "Duplicate keys when merging databoxes",
        when_no_stream="silent",
    )
    if hasattr(other, "items", ):
        other = (other, )
    for t in other:
        for key, value in t.items():
            if key in self:
                merge_strategy_func(self, key, value, stream, )
            else:
                self[key] = value
    stream._raise()

    #]


def _merge_stack(
    self,
    key: str,
    value: Any,
    stream: _wrongdoings.Stream,
    /,
) -> None:
    """
    Horizontal stack of values: time series are concatenated, lists are
    extended, non-lists are converted to lists and extended.
    """
    #[
    if isinstance(value, _series.Series, ):
        self[key] = self[key] | value
        return
    if not isinstance(self[key], list):
        self[key] = [self[key], ]
    if not isinstance(value, list):
        value = [value, ]
    self[key] += value
    #]


def _merge_replace(
    self,
    key: str,
    value: Any,
    stream: _wrongdoings.Stream,
    /,
) -> None:
    """
    """
    #[
    self[key] = value
    #]


def _merge_discard(
    self,
    key: str,
    value: Any,
    stream: _wrongdoings.Stream,
    /,
) -> None:
    """
    """
    #[
    pass
    #]


def _merge_report(
    self,
    key: str,
    value: Any,
    stream: _wrongdoings.Stream,
    /,
) -> None:
    """
    """
    #[
    stream.add(key, )
    #]


_MERGE_STRATEGY_DISPATCH = {
    "stack": _merge_stack,
    "hstack": _merge_stack,
    "replace": _merge_replace,
    "discard": _merge_discard,
    "silent": _merge_report,
    "warning": _merge_report,
    "error": _merge_report,
    "critical": _merge_report,
}




r"""
"""


#[

from __future__ import annotations

from typing import Literal
import plotly.graph_objects as _pg
import plotly.subplots as _ps

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Sequence

#]


__all__ = (
    "Chartbox",
)


BarMode = Literal["group", "relative", "stack", "overlay", ]
BarNorm = Literal["fraction", "percent", ]


class Chartbox:
    r"""
    """
    #[

    def __init__(
        self,
        *,
        subplots: tuple[int, int] | None = None,
        **kwargs,
    ) -> None:
        r"""
        """
        self.subplots = subplots if subplots is None else (int(subplots[0]), int(subplots[1]), )
        #
        if not self.subplots:
            self.figure = _pg.Figure(**kwargs, )
        else:
            self.figure = _make_subplots(
                num_rows=self.subplots[0],
                num_columns=self.subplots[1],
            )

    def show(self):
        self.figure.show()

    def add_trace(
        self,
        trace,
        subplot: tuple[int, int] | int | None = None,
        subplot_title: str | None = None,
    ) -> dict[str, int] | None:
        r"""
        """
        row_column, _ = self.resolve_subplot_reference(subplot, )
        self.figure.add_trace(trace, **row_column, )
        return row_column

    def update_layout(self, *args, **kwargs):
        self.figure.update_layout(*args, **kwargs)

    def update_traces(self, *args, **kwargs):
        self.figure.update_traces(*args, **kwargs)

    def add_line(
        self,
        *,
        subplot: tuple[int, int] | int | None = None,
        subplot_title: str | None = None,
        **kwargs,
    ) -> pg.Scatter:
        r"""
        """
        trace = new_line(**kwargs, )
        self.add_trace(trace, subplot=subplot, subplot_title=subplot_title, )
        return trace

    def add_bar(
        self,
        *,
        subplot: tuple[int, int] | int | None = None,
        subplot_title: str | None = None,
        mode: BarMode | None = None,
        norm: BarNorm | None = None,
        **kwargs,
    ) -> pg.Scatter:
        r"""
        """
        trace = new_bar(**kwargs, )
        row_column = self.add_trace(trace, subplot=subplot, subplot_title=subplot_title, )
        self.set_bar_style(mode=mode, norm=norm, )
        return trace

    def set_bar_style(
        self,
        mode: BarMode | None = None,
        norm: BarNorm | None = None,
    ) -> None:
        layout_update = {}
        if mode:
            layout_update["barmode"] = mode
        if norm:
            layout_update["barnorm"] = norm
        if layout_update:
            self.figure.update_layout(**layout_update, )

    def resolve_subplot_reference(self, *args, **kwargs, ) -> tuple[dict[str, int], int | None]:
        r"""
        """
        return resolve_subplot_reference(self.figure, *args, **kwargs, )

    def update_subplot_title(self, *args, **kwargs, ) -> _pg.layout.Annotation | None:
        r"""
        """
        return update_subplot_title(self.figure, *args, **kwargs, )

    #]


_EMPTY_SUBPLOT_TITLE = " "
_DEFAULT_VERTICAL_SPACING = 0.1
_DEFAULT_HORIZONTAL_SPACING = 0.05


def _make_subplots(
    num_rows: int,
    num_columns: int,
    #
    subplot_titles: Sequence[str] | bool | None = True,
    vertical_spacing: float | None = None,
    horizontal_spacing: float | None = None,
    figure_height: int | None = None,
    figure_title: str | None = None,
    show_legend: bool = False,
    **kwargs,
) -> _pg.Figure:
    r"""
    """
    if vertical_spacing is None:
        vertical_spacing = _DEFAULT_VERTICAL_SPACING
    if horizontal_spacing is None:
        horizontal_spacing = _DEFAULT_HORIZONTAL_SPACING
    total_numsubplots = num_rows * num_columns
    if subplot_titles is True:
        subplot_titles = (_EMPTY_SUBPLOT_TITLE, ) * total_numsubplots
    figure = _ps.make_subplots(
        rows=num_rows,
        cols=num_columns,
        subplot_titles=subplot_titles or None,
        vertical_spacing=vertical_spacing,
        horizontal_spacing=horizontal_spacing,
        **kwargs,
    )
    #
    if figure_title is not None:
        figure.update_layout(title=figure_title, )
    #
    if figure_height is not None:
        figure.update_layout(height=figure_height, )
    #
    figure.update_layout(showlegend=show_legend, )
    #
    return figure


def resolve_subplot_reference(
    figure: _pg.Figure,
    subplot: tuple[int, int] | int | None,
) -> tuple[dict[str, int], int | None]:
    r"""
    """
    if subplot is None:
        return {}, None
    rows, columns = figure._get_subplot_rows_columns()
    num_rows = len(rows)
    num_columns = len(columns)
    numsubplots = num_rows * num_columns
    if isinstance(subplot, int, ):
        if subplot < 0:
            subplot += numsubplots
        index = subplot
        row = index // num_columns
        column = index % num_columns
        row_column = {"row": row+1, "col": column+1, }
        return row_column, index,
    else:
        row = subplot[0]
        column = subplot[1]
        index = row * num_columns + column
        row_column = {"row": row+1, "col": column+1, }
        return row_column, index,


def update_subplot_title(
    figure: _pg.Figure,
    title: str,
    subplot: tuple[int, int] | int | None,
) -> _pg.layout.Annotation | None:
    r"""
    """
    row_column, index, = resolve_subplot_reference(figure, subplot, )
    annotation = next(figure.select_annotations(index, ), None, )
    if annotation:
        annotation.text = title
    return annotation


def new_line(
    *,
    y: Sequence[float],
    x: Sequence[float] | None = None,
    mode: Literal["lines", "markers", "lines+markers", ] = "lines",
    **kwargs,
) -> pg.Scatter:
    r"""
    """
    x, y, = _get_xy_data(y, x, )
    return _pg.Scatter(x=x, y=y, mode=mode, **kwargs, )


def new_bar(
    *,
    y: Sequence[float],
    x: Sequence[float] | None = None,
    **kwargs,
) -> pg.Scatter:
    r"""
    """
    x, y, = _get_xy_data(y=y, x=x, )
    trace = _pg.Bar(x=x, y=y, **kwargs, )
    return trace


def _get_xy_data(
    y: Sequence[float],
    x: Sequence[float] | None = None,
) -> tuple[tuple[float], tuple[float]]:
    r"""
    """
    y = tuple(y)
    if x is not None:
        x = tuple(x)
    else:
        x = tuple(range(len(y)))
    return x, y,




"""
"""


FUNC_STRING = """
def {n}(self, *args, **kwargs, ) -> Series:
    new = self.copy()
    out = new.{n}(*args, **kwargs, )
    if isinstance(out, tuple):
        return new, *out
    elif out is not None:
        return new, out
    else:
        return new
"""




r"""
"""


#[

from __future__ import annotations

from typing import Literal
import functools as _ft
import numpy as _np
import daqp as _qp
import ctypes as _ct

from ..dates import Period
from .main import Series

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from collections.abc import Iterable
    from numbers import Real
    from typing import Any, Callable

#]


__all__ = ("lonf", )


OrderType = Literal[1, 2]


def lonf(
    input_series,
    order: OrderType,
    smooth: Real,
    #
    span: Iterable[Period] | None = None,
) -> tuple[Any, Any]:
    r"""
................................................................................

==L1 norm based filter==

................................................................................
    """

    # Resolve periods
    periods = input_series.resolve_periods(span, )
    start_period = periods[0]
    end_period = periods[-1]
    from_until = (start_period, end_period, )
    num_periods = len(periods)

    # Set up solver function
    matrix_setup_func = _MATRIX_SETUP_DISPATCH[order]
    d, D, = matrix_setup_func(num_periods, )
    H = D @ D.T;
    A = _np.eye(num_periods-order, dtype=_ct.c_double, )
    bounds = _np.full((num_periods-order, ), smooth, dtype=_ct.c_double, );
    sense = _np.full((num_periods-order, ), 0, dtype=_ct.c_int, )
    solve = _ft.partial(_qp.solve, H=H, A=A, bupper=bounds, blower=-bounds, sense=sense, )

    # Iterate over data variants and solve the QP problem
    trend_data_variants = []
    gap_data_variants = []
    for data in input_series.iter_own_data_variants_from_until(from_until, ):
        trend_data, gap_data = _lonf_for_variant(solve, data, D, )
        trend_data_variants.append(trend_data, )
        gap_data_variants.append(gap_data, )

    # Create output series
    trend_series = Series(start=start_period, values=trend_data_variants, )
    gap_series = Series(start=start_period, values=gap_data_variants, )

    return trend_series, gap_series,


def _lonf_for_variant(
    solve: Callable,
    data: _np.ndarray,
    D: _np.ndarray,
) -> tuple[tuple[Real, ...], tuple[Real, ...]]:
    r"""
    """
    data = data.flatten()
    data.dtype = _ct.c_double
    f = -D @ data;
    x, *_ = solve(f=f, )
    gap_data = D.T @ x;
    trend_data = data - gap_data;
    trend_data = tuple(trend_data.flatten().tolist())
    gap_data = tuple(gap_data.flatten().tolist())
    return trend_data, gap_data,

#]


def _first_order_matrix_setup(
    num_periods: int,
) -> tuple[_np.ndarray, _np.ndarray]:
    r"""
    """
    order = 1
    d = _np.eye(num_periods-order, num_periods, dtype=_ct.c_double, )
    D = _np.eye(num_periods-order, num_periods, dtype=_ct.c_double, )
    D[:, 1:] = D[:, 1:] - d[:, :-1];
    return d, D,


def _second_order_matrix_setup(
    num_periods: int,
) -> tuple[_np.ndarray, _np.ndarray]:
    r"""
    """
    order = 2
    d = _np.eye(num_periods-order, num_periods, dtype=_ct.c_double, )
    D = _np.eye(num_periods-order, num_periods, dtype=_ct.c_double, )
    D[:, 1:] = D[:, 1:] - 2 * d[:, :-1] 
    D[:, 2:] = D[:, 2:] + d[:, :-2]
    return d, D,


_MATRIX_SETUP_DISPATCH = {
    1: _first_order_matrix_setup,
    2: _second_order_matrix_setup,
}




"""
Implement SimulationPlannableProtocol and SteadyPlannableProtocol
"""


#[
from __future__ import annotations

from typing import (TYPE_CHECKING, )

from .. import quantities as _quantities

from ..quantities import (
    TRANSITION_VARIABLE,
    TRANSITION_SHOCK,
    ANTICIPATED_SHOCK_VALUE,
    ENDOGENOUS_VARIABLE,
    PARAMETER,
)

if TYPE_CHECKING:
    from .main import Simultaneous
    from ..quantities import QuantityKind
#]


#-------------------------------------------------------------------------------
# Functions to be used as methods in Simultaneous class
#-------------------------------------------------------------------------------


def inlay(klass: type, ) -> type:
    r"""
    Inlay plannable protocol methods in the class
    """
    #[
    klass.get_simulation_plannable = get_simulation_plannable
    klass.get_steady_plannable = get_steady_plannable
    return klass
    #]


def get_simulation_plannable(self, **kwargs, ) -> _SimulationPlannable:
    """
    """
    return _SimulationPlannable(model=self, **kwargs, )


def get_steady_plannable(self, **kwargs, ) -> _SteadyPlannable:
    model_flags = self.resolve_flags(**kwargs, )
    return _SteadyPlannable(model=self, is_flat=model_flags.is_flat, )


#-------------------------------------------------------------------------------


class _SimulationPlannable:
    r"""
    """
    #[

    def __init__(
        self,
        model,
        **kwargs,
    ) -> None:
        r"""
        """
        #
        def get_names(kind: QuantityKind, ):
            return tuple(_quantities.generate_quantity_names_by_kind(
                model.quantities, kind=kind,
            ))
        #
        # qid_to_name = model.create_qid_to_name()
        # curr_xi_qids, *_ = model.solution_vectors.get_curr_transition_indexes()
        # can_be_exogenized = tuple( qid_to_name[qid] for qid in curr_xi_qids )
        #
        can_be_exogenized = get_names(TRANSITION_VARIABLE, )
        self.can_be_exogenized_anticipated = can_be_exogenized
        self.can_be_exogenized_unanticipated = can_be_exogenized
        self.can_be_endogenized_unanticipated = get_names(TRANSITION_SHOCK, )
        self.can_be_endogenized_anticipated = get_names(ANTICIPATED_SHOCK_VALUE, )

    #]


class _SteadyPlannable:
    r"""
    """
    #[

    def __init__(
        self,
        model: Simultaneous,
        is_flat: bool,
        **kwargs,
    ) -> None:
        r"""
        """
        def get_names(kind: QuantityKind, ):
            return tuple(_quantities.generate_quantity_names_by_kind(
                model.quantities, kind=kind,
            ))
        #
        self.can_be_exogenized = get_names(ENDOGENOUS_VARIABLE, )
        # self.can_be_exogenized = get_names(TRANSITION_VARIABLE, )
        self.can_be_endogenized = get_names(PARAMETER, )
        self.can_be_fixed_level = self.can_be_exogenized
        self.can_be_fixed_change = ()
        if not is_flat:
            self.can_be_fixed_change = self.can_be_exogenized

    #]




"""
"""


#[

from __future__ import annotations

from typing import (Self, Any, )
from numbers import (Real, )
import numpy as _np
import itertools as _it
import functools as _ft
import warnings as _wa
import documark as _dm

from .. import wrongdoings as _wrongdoings
from .. import has_variants as _has_variants
from ..databoxes.main import Databox
from ..plans.simulation_plans import SimulationPlan
from ..plans.transforms import PlanTransform
from ..explanatories import main as _explanatories
from ..dataslates.main import Dataslate

#]


#-------------------------------------------------------------------------------
# Functions to be used as methods in Simultaneous class
#-------------------------------------------------------------------------------


def inlay(klass, ) -> type:
    """
    """
    klass.simulate = simulate
    return klass


@_dm.reference(category="simulation", )
def simulate(
    self,
    input_db: Databox,
    span: Iterable[Dater],
    *,
    plan: SimulationPlan | None = None,
    prepend_input: bool = True,
    target_db: Databox | None = None,
    when_nonfinite = None,
    when_simulates_nan: Literal["error", "warning", "silent", ] = "warning",
    execution_order: Literal["dates_equations", "equations_dates", ] = "dates_equations",
    num_variants: int | None = None,
    remove_initial: bool = True,
    remove_terminal: bool = True,
    shocks_from_data: bool = True,
    parameters_from_data: bool = False,
    catch_warnings: bool = False,
    unpack_singleton: bool = True,
    return_info: bool = False,
    method: Literal["sequential", ] = "sequential",
) -> tuple[Databox, dict[str, Any]]:
    r"""
................................................................................

==Simulate sequential model==

Simulate a `Sequential` model, `self`, on a time `span`, period by period,
equation by equation. The `simulate` function does not reorder the
equations; if needed, this must be done by running `reorder` before
simulating the model.


```
out_db = self.simulate(
input_db,
simulation_span,
*,
plan=None,
execution_order="dates_equations",
prepend_input=True,
target_db=None,
when_simulates_nan="warning",
num_variants=None,
remove_initial=True,
remove_terminal=True,
return_info=False,
)
```

```
out_db, info = self.simulate(
...,
return_info=True,
...,
)
```


### Input arguments ###


???+ input "self"
`Sequential` model that will be simulated.

???+ input "input_db"
Input databox (a `Databox` object) with all the necessary initial
conditions (initial lags) for the LHS variables, and all the values for
the RHS variables needed to simulate `self` on the time `span`.

???+ input "plan"
`PlanSimulate` object with a simulation plan, i.e. information about
which LHS variables to exogenize at which dates. If `plan=None`, no
simulation plan is imposed on the simulation.

???+ input "simulation_span"
[Time span](../data_management/spans.md) for the simulation; the time
span needs to go forward and have a one-period step.

???+ input "plan"
[Simulation plan](simulation_plans.md) for the simulation specifying the
exogenized data points. If `None`, no simulation plan is imposed.

???+ input "prepend_input"
If `True`, the input time series observations are prepended to the results.

???+ input "target_db"
Custom databox to which the simulated time series will be added. If
`None`, a new databox is created.

???+ input "when_simulates_nan"
Action to take when a simulated data point is non-finite (`nan` or `inf` or `-inf`). The options are

* `"error"`: raise an error,
* `"warning"`: log a warning,
* `"silent"`: do nothing.

???+ input "execution_order"
Order in which the model equations and simulation periods are executed. The options are

* `"dates_equations"`: all equations for the first period, all equations for the second period, …
* `"equations_dates"`: all periods for the first equation, all periods for the second equation, …

???+ input "num_variants"
Number of variants to simulate. If `None`, the number of variants is
determined by the number of variants in the `self` model.

???+ input "remove_initial"
If `True`, remove the initial condition data, i.e. all lags before the
start of the simulation span.

???+ input "remove_terminal"
If `True`, remove the terminal condition data, i.e. all leads after the
end of the simulation span.


### Returns ###


???+ returns "out_db"
Output databox with the simulated time series for the LHS variables.

???+ returns "info"
(Only returned if `return_info=True which is not the default behavior)
Dictionary with information about the simulation; `info` contains the
following items:

| Key | Description
|-----|-------------
| `"method"` | Simulation method used
| `"execution_order"` | Execution order of the equations and periods


................................................................................
    """
    # Legacy
    if when_nonfinite is not None:
        when_simulates_nan = when_nonfinite
    #
    num_variants \
        = self.resolve_num_variants_in_context(num_variants, )
    #
    base_dates = tuple(span, )
    #
    extra_databox_names = None
    if plan is not None:
        plan.check_consistency(self, base_dates, )
        extra_databox_names = plan.get_databox_names()
    #
    slatable = self.slatable_for_simulate(
        shocks_from_data=shocks_from_data,
        parameters_from_data=parameters_from_data,
    )
    dataslate = Dataslate.from_databox_for_slatable(
        slatable, input_db, base_dates,
        num_variants=self.num_variants,
        extra_databox_names=extra_databox_names,
    )
    #
    zipped = zip(
        range(num_variants, ),
        self.iter_variants(),
        dataslate.iter_variants(),
    )
    #
    #
    #=======================================================================
    # Main loop over variants
    out_info = []
    simulate_method = _SIMULATION_METHOD_DISPATCH[method]
    for vid, model_v, dataslate_v in zipped:
        info_v = simulate_method(
            model_v, dataslate_v, plan, vid,
            when_simulates_nan=when_simulates_nan,
            execution_order=execution_order,
            catch_warnings=catch_warnings,
        )
        out_info.append(info_v, )
    #=======================================================================
    #
    #
    # Remove initial and terminal condition data (all lags and leads
    # before and after the simulation span)
    if remove_terminal:
        dataslate.remove_terminal()
    if remove_initial:
        dataslate.remove_initial()
    #
    # Convert all variants of the dataslate to a databox
    out_db = dataslate.to_databox()
    if prepend_input:
        out_db.prepend(input_db, base_dates[0]-1, )
    #
    # Add to custom databox
    if target_db is not None:
        out_db = target_db | out_db
    #
    if return_info:
        out_info = _has_variants.unpack_singleton(
            out_info, self.is_singleton,
            unpack_singleton=unpack_singleton,
        )
        return out_db, out_info
    else:
        return out_db


#-------------------------------------------------------------------------------


def _simulate_v(
    self,
    ds: Dataslate,
    plan: SimulationPlan | None,
    vid: int,
    *,
    when_simulates_nan,
    execution_order,
    catch_warnings,
) -> dict[str, Any]:
    r"""
    """
    when_nonfinite_stream = \
        _wrongdoings.STREAM_FACTORY[when_simulates_nan] \
        ("These simulated data point(s) are nan or inf:", )
    #
    name_to_row = ds.create_name_to_row()
    base_columns = ds.base_columns
    first_base_column = base_columns[0] if base_columns else None
    working_data = ds.get_data_variant(0, )
    columns_dates = ( (c, ds.periods[c]) for c in base_columns )
    iterator_creator = _CREATE_EXECUTION_ITERATOR[execution_order]
    #
    columns_dates_equations = iterator_creator(
        columns_dates,
        self.iter_equations(),
    )
    #
    detect_exogenized = _ft.partial(
        _detect_exogenized,
        name_to_row=name_to_row,
        data=working_data,
    )
    #
    info = {
        "method": "sequential",
        "execution_order": execution_order,
    }
    #
    if catch_warnings:
        _wa.simplefilter("error", )
    for (column, date), equation in columns_dates_equations:
        lhs_date_str = f"{equation.lhs_name}[{date}]"
        residual_date_str = f"{equation.residual_name}[{date}]"
        transform = _get_transform(plan, equation, date)
        try:
            implied_value = detect_exogenized(equation.lhs_name, transform, column, )
            simulation_func = (
                equation.simulate
                if implied_value is None
                else equation.exogenize
            )
            info_eq = simulation_func(working_data, column, implied_value, )
        except Exception as exc:
            message = (
                f"Error when simulating {lhs_date_str}"
                f"\nDirect cause: {str(exc)}"
            )
            raise _wrongdoings.IrisPieCritical(message, ) from exc
        #
        _catch_nonfinite(
            when_nonfinite_stream,
            info_eq["is_finite"],
            info_eq["simulated_name"],
            info_eq["simulated_value"],
            date,
        )
    when_nonfinite_stream._raise()
    _wa.simplefilter("default", )
    #
    return info


def _get_transform(
    plan: SimulationPlan | None,
    equation: _explanatories.Explanatory,
    date: _dates.Dater,
) -> PlanTransform | None:
    """
    """
    return (
        plan.get_exogenized_point(equation.lhs_name, date, )
        if plan and not equation.is_identity
        else None
    )


def _detect_exogenized(
    lhs_name: str,
    transform: PlanTransform | None,
    data_column: int,
    *,
    data: _np.ndarray,
    name_to_row: dict[str, int],
) -> Real | None:
    """
    """
    #[
    if transform is None:
        return None
    #
    lhs_name_row = name_to_row[lhs_name]
    values_before = data[lhs_name_row, :data_column]
    values_after_inclusive = data[lhs_name_row, data_column:]
    #
    transform_name = transform.resolve_databox_name(lhs_name, )
    transform_row = name_to_row.get(transform_name, None, )
    transform_values_after = (
        data[transform_row, data_column:]
        if transform_row is not None else None
    )
    implied_value = transform.eval_exogenized(transform_values_after, values_before, values_after_inclusive, )
    if transform.when_data and _np.isnan(implied_value):
        return None
    return implied_value
    #]


def _catch_nonfinite(
    stream: _wrongdoings.Stream,
    is_finite: _np.ndarray,
    simulated_name: str,
    simulated_value: Real | _np.ndarray,
    date: Dater,
) -> None:
    """
    """
    #[
    if is_finite.all():
        return
    simulate_value = (
        simulated_value.tolist()
        if isinstance(simulated_value, _np.ndarray)
        else simulated_value
    )
    message = f"{simulated_name}[{date}]={simulated_value}"
    stream.add(message, )
    #]


def _iter_dates_equations(columns_dates, equations, ) -> Iterator[tuple[int, _dates.Dater], _explanatories.Explanatory]:
    return _it.product(columns_dates, equations, )


def _iter_equations_dates(columns_dates, equations, ) -> Iterator[tuple[int, _dates.Dater], _explanatories.Explanatory]:
    return _swap_product(_it.product(equations, columns_dates, ), )


def _swap_product(iterator: Iterable[tuple[Any, Any]], ) -> Iterable[tuple[Any, Any]]:
    return ((b, a) for a, b in iterator)


_SIMULATION_METHOD_DISPATCH = {
    "sequential": _simulate_v,
}


_CREATE_EXECUTION_ITERATOR = {
    "dates_equations": _iter_dates_equations,
    "equations_dates": _iter_equations_dates,
}




r"""
Description mixin
"""


#[

from __future__ import annotations

from typing import Protocol
import documark as _dm

#]


class DescriptionableProtocol(Protocol, ):
    r"""
    """
    #[

    __description__: str | None = None

    #]


class DescriptionMixin:
    r"""
    """
    #[

    @_dm.reference(category="information", )
    def get_description(
        self: DescriptionableProtocol,
    ) -> str:
        r"""
................................................................................


==Get description attached to an object==

    description = self.get_description()


### Input arguments ###

???+ input "self"
    An object from which to get the description.


### Returns ###

???+ returns "description"
    The description attached to the object.


................................................................................
        """
        return str(self.__description__ or "")

    @_dm.reference(category="information", )
    def set_description(
            self: DescriptionableProtocol,
            description: str,
        ) -> None:
        r"""
................................................................................


==Set the description for an object==

    self.set_description(
        description,
    )


### Input arguments ###

???+ input "self"
    An Iris Pie object to which to attach the description.


???+ input "description"
    The description to attach to the Iris Pie object.


### Returns ###

This method modifies the Iris Pie object in place and returns `None`.


................................................................................
        """
        self.__description__ = str(description or "")

    #]




r"""
Kalman filter inlay
"""


#[

from __future__ import annotations

import numpy as _np
import documark as _dm

from ..fords import kalmans as _kalmans

#]


# TODO: Create Kalmanable protocol


class Inlay:
    r"""
    """
    #[

    @_dm.reference(category="filtering", )
    def kalman_filter(self, *args, **kwargs, ):
        r"""
        """
        return _kalmans.kalman_filter(
            self,
            *args,
            generate_period_system=_generate_period_system,
            generate_period_data=_generate_period_data,
            **kwargs,
        )

    @_dm.reference(category="filtering", )
    def neg_log_likelihood(self, *args, **kwargs, ):
        r"""
        """
        return _kalmans.neg_log_likelihood(
            self,
            *args,
            generate_period_system=_generate_period_system,
            generate_period_data=_generate_period_data,
            **kwargs,
        )

    #]


def _generate_period_system(
    t: int,
    #
    solution_v: Solution,
    y1_array: _np.ndarray,
    std_u_array: _np.ndarray,
    std_w_array: _np.ndarray,
    all_v_impact: Sequence[_np.ndarray | None] | None,
) -> tuple[_np.ndarray, ...]:
    """
    """
    #[
    T = solution_v.Ta
    P = solution_v.Pa
    K = solution_v.Ka
    inx_y = ~_np.isnan(y1_array[:, t], )
    Z = solution_v.Za[inx_y, :]
    H = solution_v.H[inx_y, :]
    D = solution_v.D[inx_y]
    U = solution_v.Ua
    cov_u = _np.diag(std_u_array[:, t]**2, )
    cov_w = _np.diag(std_w_array[:, t]**2, )
    v_impact = all_v_impact[t] if all_v_impact is not None else None
    return T, P, K, Z, H, D, cov_u, cov_w, v_impact, U,
    #]


def _generate_period_data(
    t,
    y_array: _np.ndarray,
    u_array: _np.ndarray,
    v_array: _np.ndarray,
    w_array: _np.ndarray,
) -> tuple[_np.ndarray, ...]:
    """
    """
    #[
    inx_y = ~_np.isnan(y_array[:, t], )
    y = y_array[inx_y, t]
    u = u_array[:, t]
    v = v_array[:, t]
    w = w_array[:, t]
    return y, u, v, w, inx_y.tolist(),
    #]




"""
Simultaneous models
"""


#[

from __future__ import annotations

from typing import Self, Any, TypeAlias, Literal
from types import EllipsisType
from collections.abc import Iterable, Iterator, Callable
from numbers import Number
import numpy as _np
import itertools as _it
import functools as _ft
import documark as _dm

from .. import has_invariant as _has_invariant
from .. import has_variants as _has_variants
from .. import equations as _equations
from .. import quantities as _quantities
from .. import portables as _portables
from ..quantities import QuantityKind, Quantity
from ..sources import ModelSource
from .. import sources as _sources
from .. import dates as _dates
from .. import wrongdoings as _wrongdoings
from ..conveniences import iterators as _iterators
from ..parsers import common as _pc
from ..databoxes import main as _databoxes

from ..fords import solutions as _solutions
from ..fords import steadiers as _fs
from ..fords import descriptors as _descriptors
from ..fords import systems as _systems
from ..fords import std_simulators as _std_simulators

from ._invariants import Invariant
from ._variants import Variant

from . import _assigns
from . import _covariances
from . import _flags
from . import _get
from . import _io
from . import _kalmans
from . import _logly
from . import _plannable_protocols
from . import _pretty
from . import _simulate
from . import _slatable_protocols
from . import _steady
from . import _steady_boxable_protocols
from . import _tolerance

#]


__all__ = [
    "Simultaneous",
    "Model",
]


@_pretty.inlay
@_steady_boxable_protocols.inlay
@_slatable_protocols.inlay
@_steady.inlay
@_plannable_protocols.inlay
@_dm.reference(
    path=("structural_models", "simultaneous.md", ),
    categories={
        "constructor": "Creating new Simultaneous models",
        "information": "Getting information about Simultaneous models",
        "filtering": "Applying structural filters on Simultaneous models",
        "parameters": "Manipulating Simultaneous model parameters",
        "serialize": "Serializing, saving and loading Simultaneous models",
        "property": None,
    },
)
class Simultaneous(
    _has_invariant.Mixin,
    _has_variants.Mixin,
    _std_simulators.Mixin,
    #
    _assigns.Inlay,
    _simulate.Inlay,
    _kalmans.Inlay,
    _logly.Inlay,
    _get.Inlay,
    _covariances.Inlay,
    _tolerance.Inlay,
    _io.Inlay,
):
    """
................................................................................

`Simultaneous` models
======================


................................................................................
    """
    #[

    __slots__ = (
        "_invariant",
        "_variants",
    )

    def __init__(
        self,
        invariant: Invariant | None = None,
        variants: Iterable[Variant] | None = None,
    ) -> None:
        r"""
        """
        self._invariant = invariant
        self._variants = list(variants) if variants else []

    @classmethod
    @_dm.reference(category="constructor", call_name="Simultaneous.from_file", )
    def from_file(klass, *args, **kwargs, ) -> _sources.SourceMixinProtocol:
        r"""
················································································

==Create `Simultaneous` model object from source file or files==

Read and parse one or more source files specified by `file_names` (a string
or a list of strings) with model source code, and create a `Simultaneous`
model object.

```
self = Simultaneous.from_file(
    file_names,
    #
    context=None,
    description="",
)
```


### Input arguments ###

???+ input "file_names"
    The name of the model source file from which the `Simultaneous` model object
    will be created, or a list of file names; if multiple file names are
    specified, they will all combined together in the given order.

???+ input "context"
    Dictionary supplying the values used in preparsing commands, and the
    definition of non-standard functions used in the equations.

???+ input "description"
    Desscription of the model specified as a text string.


### Returns ###

???+ returns "self"
    `Simultaneous` model object created from the `file_names`.

················································································
        """
        return _sources.from_file(klass, *args, **kwargs, )

    @classmethod
    @_dm.reference(category="constructor", call_name="Simultaneous.from_string",)
    def from_string(klass, *args, **kwargs, ) -> _sources.SourceMixinProtocol:
        """
················································································

==Create `Simultaneous` model from string==

Read and parse a text `string` with a model source code, and create a
`Simultaneous` model object. Otherwise, this function behaves the same way as
[`Simultaneous.from_file`](#simultaneousfrom_file).

```
self = Simultaneous.from_string(
    string,
    #
    context=None,
    description="",
)
```


### Input arguments ###


???+ input "string"

    Text string from which the `Simultaneous` model object will be created.

See [`Simultaneous.from_file`](#simultaneousfrom_file) for other input arguments.


### Returns ###

See [`Simultaneous.from_file`](simultaneousfrom_file) for return values.

················································································
        """
        return _sources.from_string(klass, *args, **kwargs, )

    def copy(self, ) -> Self:
        """
        Create a quasi-deep copy of this model
        """
        new = type(self)()
        new._invariant = self._invariant.copy()
        new._variants = [ variant.copy() for variant in self._variants ]
        return new

    def __getitem__(
        self,
        request: str | int,
    ):
        """
        Implement self[i] for variants and self[name] for quantitie values
        """
        if isinstance(request, str):
            return self.get_value(request, )
        else:
            return self.get_variant(request, )

    def __setitem__(
        self,
        reference: str,
        value: Any,
    ) -> None:
        r"""
        """
        if isinstance(reference, str):
            self._assign({reference: value}, )
        else:
            raise NotImplementedError("Assigning model variants not implemented yet", )

    def __repr__(self, ) -> str:
        """
        """
        return "\n".join((
            f"",
            f"<{self.__class__.__name__} at {id(self):#x}>",
            f"[Description: \"{self.get_description()}\"]",
            f"[Num variants: {self.num_variants}]",
            f"[Num transition: {self.num_transition_equations}]",
            f"[Num measurement: {self.num_measurement_equations}]",
            f"[Max lag: t{self.max_lag:+g}]",
            f"[Max lead: t{self.max_lead:+g}]",
            f"",
        ))

    def get_value(
        self,
        name: str,
    ) -> Any:
        """
        """
        quantities = self._invariant.quantities
        qids, invalid_names = _quantities.lookup_qids_by_name(quantities, (name, ), )
        if invalid_names:
            raise _wrongdoings.IrisPieCritical(f"Invalid model name \"{invalid_names[0]}\"", )
        return _has_variants.unpack_singleton(
            self._get_values_as_dict("levels", qids, )[name],
            self.is_singleton,
        )

    def change_logly(
        self,
        new_logly: bool,
        some_names: Iterable[str] | None = None,
        /
    ) -> None:
        """
        Change the log-status of some model quantities
        """
        some_names = set(some_names) if some_names else None
        qids = [
            qty.id
            for qty in self._invariant.quantities
            if qty.logly is not None and (some_names is None or qty.human in some_names)
        ]
        self._invariant.quantities = _quantities.change_logly(self._invariant.quantities, new_logly, qids)

    @property
    @_dm.reference(category="property", )
    def is_linear(self, ) -> bool:
        """==True for models declared as linear=="""
        return self._invariant._flags.is_linear

    @property
    @_dm.reference(category="property", )
    def is_flat(self, ) -> bool:
        """==True for models declared as flat=="""
        return self._invariant._flags.is_flat

    @property
    @_dm.reference(category="property", )
    def is_deterministic(self, ) -> bool:
        """==True for models declared as deterministic=="""
        return self._invariant._flags.is_deterministic

    @property
    @_dm.reference(category="property", )
    def num_transition_equations(self, ) -> int:
        """==Number of transition equations=="""
        return self._invariant.num_transition_equations

    @property
    @_dm.reference(category="property", )
    def num_measurement_equations(self, ) -> int:
        """==Number of measurement equations=="""
        return self._invariant.num_measurement_equations

    @property
    @_dm.reference(category="property", )
    def max_lag(self, ) -> int:
        """==Maximul lag in the model (negative or zero)=="""
        return self._invariant._min_shift

    @property
    @_dm.reference(category="property", )
    def max_lead(self, ) -> int:
        """==Maximul lead in the model (positive or zero)=="""
        return self._invariant._max_shift

    @property
    def solution_vectors(self, ) -> _descriptors.SolutionVectors:
        """
        """
        return self._invariant.dynamic_descriptor.solution_vectors

    @property
    def quantities(self, ) -> tuple[Quantity]:
        """==Tuple of model quantities=="""
        return self._invariant.quantities

    @property
    def shock_qid_to_std_qid(self, ) -> dict[int, int]:
        """==Dictionary mapping shock quantity id to standard deviation quantity id=="""
        return self._invariant.shock_qid_to_std_qid

    def create_steady_array(
        self,
        variant: Variant | None = None,
        **kwargs,
    ) -> _np.ndarray:
        """
        """
        qid_to_logly = self.create_qid_to_logly()
        if variant is None:
            variant = self._variants[0]
        return variant.create_steady_array(qid_to_logly, **kwargs, )

    def create_zero_array(
        self,
        variant: Variant | None = None,
        **kwargs,
    ) -> _np.ndarray:
        """
        """
        qid_to_logly = self.create_qid_to_logly()
        if variant is None:
            variant = self._variants[0]
        return variant.create_zero_array(qid_to_logly, **kwargs, )

    def create_some_array(
        self,
        deviation: bool,
        **kwargs,
    ) -> _np.ndarray:
        """
        """
        if not deviation:
            return self.create_steady_array(**kwargs, )
        else:
            return self.create_zero_array(**kwargs, )

    def _enforce_assignment_rules(self, variant, ) -> None:
        """
        """
        #
        # Reset levels of shocks to zero
        name_to_qid = self.create_name_to_qid()
        shock_qids = _quantities.generate_qids_by_kind(
            self._invariant.quantities, _quantities.QuantityKind.ANY_SHOCK_OR_SHOCK_VALUE,
        )
        zero_shocks = { i: 0 for i in shock_qids }
        variant.update_values_from_dict(zero_shocks, )
        #
        # Remove changes from quantities that are not loggable
        non_loggables = _quantities.generate_qids_by_kind(
            self._invariant.quantities, ~QuantityKind.LOGGABLE_VARIABLE,
        )
        assign_non_loggables = { qid: (..., _np.nan, ) for qid in non_loggables }
        variant.update_values_from_dict(assign_non_loggables, )

    def systemize(
        self,
        unpack_singleton: bool = True,
        **kwargs,
    ) -> Iterable[_systems.System]:
        """
        Create unsolved first-order system for each variant
        """
        model_flags = self.resolve_flags(**kwargs, )
        system = [
            self._systemize(variant, self._invariant.dynamic_descriptor, model_flags, )
            for variant in self._variants
        ]
        return _has_variants.unpack_singleton(
            system, self.is_singleton,
            unpack_singleton=unpack_singleton,
        )

    def _systemize(
        self,
        variant: Variant,
        descriptor: _descriptors.Descriptor,
        model_flags: flags.Flags,
    ) -> _systems.System:
        """
        Create unsolved first-order system for one variant
        """
        min_shift = self._invariant._min_shift
        max_shift = self._invariant._max_shift
        num_columns = -min_shift + 1 + max_shift
        qid_to_logly = self.create_qid_to_logly()
        #
        if model_flags.is_linear:
            data_array = variant.create_zero_array(qid_to_logly, num_columns=num_columns, shift_in_first_column=min_shift, )
            data_array_lagged = None
        else:
            data_array = variant.create_steady_array(qid_to_logly, num_columns=num_columns, shift_in_first_column=min_shift, )
            data_array_lagged = variant.create_steady_array(qid_to_logly, num_columns=num_columns, shift_in_first_column=min_shift-1, )
        #
        column_offset = -min_shift
        return _systems.System(
            descriptor, data_array,
            model_flags, data_array_lagged, column_offset,
        )

    def solve_first_order(
        self,
        clip_small: bool = False,
        return_info: bool = False,
        unpack_singleton: bool = True,
        tolerance: float | None = None,
        **kwargs,
    ) -> dict[str, Any]:
        r"""
        Calculate first-order solution for each variant within this model
        """
        model_flags = self.resolve_flags(**kwargs, )
        tolerance = tolerance or self.get_tolerance("eigenvalue", )
        out_info = [
            self._solve_variant(
                self_v,
                vid,
                model_flags,
                tolerance=tolerance,
                clip_small=clip_small,
            )
            for vid, self_v in enumerate(self._variants, )
        ]
        if return_info:
            out_info = _has_variants.unpack_singleton(
                out_info, self.is_singleton,
                unpack_singleton=unpack_singleton,
            )
            return out_info
        else:
            return

    solve = solve_first_order

    def _solve_variant(
        self,
        variant: Variant,
        vid: int,
        model_flags: flags.Flags,
        tolerance: float,
        clip_small: bool,
    ) -> None:
        r"""
        Calculate first-order solution for one variant of this model
        """
        variant_header = f"[Variant {vid}]"
        system = self._systemize(
            variant,
            self._invariant.dynamic_descriptor,
            model_flags,
        )
        try:
            variant.solution = _solutions.Solution.from_system(
                self._invariant.dynamic_descriptor,
                system,
                clip_small=clip_small,
                tolerance=tolerance,
            )
        except _solutions.UnitRootException:
            raise _wrongdoings.IrisPieCritical(
                f"{variant_header} Inconsistency in classification of unit roots; "
                "modify (increase) the tolerance level",
            )
        info = {}
        #
        return info

    def _choose_plain_equator(
        self,
        equation_switch: Literal["dynamic", "steady", ],
    ) -> Callable | None:
        """
        """
        match equation_switch:
            case "dynamic":
                return self._invariant._plain_dynamic_equator
            case "steady":
                return self._invariant._plain_steady_equator

    def reset_stds(self, ) -> None:
        """
        Initialize standard deviations of shocks to default values
        """
        std_names = _quantities.generate_quantity_names_by_kind(
            self._invariant.quantities,
            _quantities.QuantityKind.ANY_STD,
        )
        dict_to_assign = {
            k: self._invariant._default_std
            for k in std_names
        }
        self.assign(**dict_to_assign, )

    @classmethod
    def from_source(
        klass,
        source: ModelSource,
        **kwargs,
    ) -> Self:
        """
        """
        self = klass()
        self._invariant = Invariant.from_source(source, **kwargs, )
        self._initialize_variants_from_invariant()
        return self

    def _initialize_variants_from_invariant(self, ) -> None:
        initial_variant = Variant.from_source(
            self._invariant.quantities,
            self._invariant._flags.is_flat,
        )
        self._variants = [ initial_variant ]
        self.reset_stds()
        self._enforce_assignment_rules(self._variants[0], )

    def get_context(self, ) -> dict[str, Any]:
        """
        """
        return self._invariant._context

    def resolve_flags(self, **kwargs, ) -> _flags.Flags:
        return _flags.Flags.update_from_kwargs(self.get_flags(), **kwargs)

    def __getstate__(self, ):
        """
        """
        return {
            "_invariant": self._invariant,
            "_variants": self._variants,
        }

    def __setstate__(self, state, ):
        """
        """
        self._invariant = state["_invariant"]
        self._variants = state["_variants"]

    @_dm.reference(category="serialize", )
    def to_portable(self, ) -> dict[str, Any]:
        r"""
................................................................................

==Serialize `Simultaneous` model to portable dictionary==

Convert a `Simultaneous` model object to a dictionary of primitive values that
can be saved to a JSON file or transmitted over the network. The structure of
the portable dictionary is described below.

```
portable = self.to_portable()
```


### Input arguments ###

???+ input "self"
    `Simultaneous` model object to be serialized.


### Returns ###

???+ returns "portable"
    A JSON-serializable dictionary with the structure described below.


### Details ###


???+ abstract "Structure of the portable dictionary, format 0.3.0:"

    The portable dictionary has the following structure:

    ```
        {
            "portable_format": "0.3.0",
            "source": {
                "description": <str>,
                "flags": {
                    "is_linear: <bool>,
                    "is_flat": <bool>,
                    "is_deterministic": <bool>
                },
                "quantities": [ <QUANTITY>, ... ],
                "equations": [ <EQUATION>, ... ],
                "context": {},
            "variants": [ <VARIANT>, ... ],
        }
    ```

    The meaning of the `flags` is as follows:

    * `is_linear` is `True` if the model has been created with the
    `linear=True`, and the calculation of first-order solution matrices is
    expected to be independent of the model steady state. * `is_flat` is `True`

    * `is_flat` is `True` if the model has been created with the `flat=True`,
    and the calculation of the model steady state is done assuming that no model
    variable is growing over time in steady state.

    * `is_deterministic` is `True` if the model has been created with the
    `deterministic=True`, and the model is expected to have no stochastic
    shocks; all shocks are assumed to be deterministic add-factors, and there
    are no `std` parameter created.

    The `quantities` is a list of quantities, with each `<QUANTITY>` described
    as a five-element tuple:

    ```
    [ <KIND>, <NAME>, <LOGLY>, <DESCRIPTION>, <ATTRIBUTES> ]
    ```

    where

    * `<KIND>` is a string representing the kind of the quantity,

    * `<NAME>` is a string with the name of the quantity,

    * `<LOGLY>` is a boolean or `None` indicating whether the quantity is
    declared as a log variable `None` if irrelevant for the respective kind of variables),

    * `<DESCRIPTION>` is a string with the description of the quantity,

    * `<ATTRIBUTES>` is a string containing all attributes separated by a white space.


    The kinds of variables are coded as follows:

    | Kind of variable | Code |
    |-------------------|------|
    | Transition variable | `#x` |
    | Measurement variable | `#y` |
    | Unanticipated shock | `#u` |
    | Measurement shock | `#w` |
    | Parameter | `#p` |
    | Exogenous variable | `#z` |


    The `equations` is a list of equations, with each `<EQUATION>` described as
    a five-tuple:

    ```
    [ <KIND>, <DYNAMIC>, <STEADY>, <DESCRIPTION>, <ATTRIBUTES> ]
    ```

    where
    * `<KIND>` is a string representing the kind of the equation,

    * `<DYNAMIC>` is a string with the dynamic variant of the equation,

    * `<STEADY>` is a string with the steady-state variant of the equation,

    * `<DESCRIPTION>` is a string with the description of the equation,

    * `<ATTRIBUTES>` is a string containing all attributes separated by a white space.


    The kinds of equations are coded as follows:

    | Kind of equation | Code |
    |-------------------|------|
    | Transition equation | `#T` |
    | Measurement equation | `#M` |
    | Steady autovalues | `#A` |


    The `variants` is a list of parameter (and steady-state) variants, with each
    `<VARIANT>` being a dictionary with its keys corresponding to the names of
    the model quantities (including `std_` names for the standard deviations of
    shocks), and the values being two-tuples consisting of the level and the
    changes of the respective quantity; the change is `None` whenever irrelevant
    for the respective kind of quantity.

................................................................................
        """
        qid_to_name = self.create_qid_to_name()
        return {
            "portable_format": _portables.CURRENT_PORTABLE_FORMAT,
            "source": self._invariant.to_portable(),
            "variants": [ v.to_portable(qid_to_name, ) for v in self._variants ],
        }

    @classmethod
    @_dm.reference(category="constructor", )
    def from_portable(klass, portable, ) -> Self:
        r"""
................................................................................

==Create `Simultaneous` model from portable dictionary==

Create a `Simultaneous` model object from a portable dictionary. See
`Simultaneous.to_portable` for details on the structure of the portable
dictionary.

```
self = Simultaneous.from_portable(portable)
```


### Input arguments ###

???+ input "portable"
    A dictionary with the structure described in the documentation of
    `Simultaneous.to_portable`.


### Returns ###

???+ returns "self"
    A new `Simultaneous` model object created from the portable dictionary.

................................................................................
        """
        _portables.validate_portable_format(portable["portable_format"], )
        #
        self = klass()
        self._invariant = Invariant.from_portable(portable["source"], )
        self._initialize_variants_from_invariant()
        num_variants = len(portable["variants"])
        self.alter_num_variants(num_variants, )
        for self_v, portable_v in zip(self.iter_own_variants(), portable["variants"], ):
            self_v.assign_strict(portable_v, )
        return self

    #]


Model = Simultaneous




"""
Input-output operations
"""


#[

from __future__ import annotations

import json as _js
import pickle as _pk
import dill as _dl
import documark as _dm

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Self, Any

#]


class Inlay:
    """
    """
    #[

    def to_pickle_bytes(self, **kwargs, ) -> bytes:
        """
        """
        return _pk.dumps(self, **kwargs, )

    def to_pickle_file(self, file_name: str, **kwargs, ) -> None:
        """
        """
        with open(file_name, "wb", ) as file:
            _pk.dump(self, file, **kwargs, )

    def to_dill_bytes(self, **kwargs, ) -> bytes:
        """
        """
        return _dl.dumps(self, **kwargs, )

    def to_dill_file(self, file_name: str, **kwargs, ) -> None:
        """
        """
        with open(file_name, "wb", ) as file:
            _dl.dump(self, file, **kwargs, )

    @classmethod
    def from_pickle_file(klass, file_name: str, **kwargs, ) -> Self:
        """
        """
        with open(file_name, "rb", ) as file:
            return _pk.load(file, **kwargs, )

    @classmethod
    def from_dill_file(klass, file_name: str, **kwargs, ) -> Self:
        """
        """
        with open(file_name, "rb", ) as file:
            return _dl.load(file, **kwargs, )

    @_dm.reference(category="serialize", )
    def to_portable_file(
        self,
        file_name: str,
        json_settings: None | dict[str, Any] = None,
        **kwargs,
    ) -> None:
        r"""
................................................................................

==Write a `Simultaneous` object to a portable JSON file==

Convert the `Simultaneous` object to a portable dictionary and write it to a
JSON file. See [`Simultaneous.to_portable`](#to_portable) for details on the
conversion to a portable dictionary.


```
self.to_portable_file(
    file_name,
    json_settings=None,
)
```


### Input arguments ###

???+ input "file_name"
    Filename under which to save the JSON file.

???+ input "json_settings"
    Optional settings to pass to `json.dump`.


### Returns ###

The method returns `None`.

................................................................................
        """
        portable = self.to_portable(**kwargs, )
        with open(file_name, "wt", ) as file:
            _js.dump(portable, file, **(json_settings or {}), )

    @classmethod
    @_dm.reference(category="constructor", )
    def from_portable_file(
        klass,
        file_name: str,
        json_settings: None | dict[str, Any] = None,
        **kwargs,
    ) -> Self:
        r"""
................................................................................

==Read a `Simultaneous` object from a portable JSON file==

Read a JSON file and convert the contents to a `Simultaneous` object. See
[`Simultaneous.from_portable`](#from_portable) for details on the conversion
from a portable dictionary.

```
self = Simultaneous.from_portable_file(
    file_name,
    json_settings=None,
)
```


### Input arguments ###

???+ input "file_name"
    Filename from which to read the JSON file.

???+ input "json_settings"
    Optional settings to pass to `json.load`.


### Returns ###

???+ returns "self"
    A new `Simultaneous` object created from the contents of the JSON file.

................................................................................

        """
        with open(file_name, "rt", ) as file:
            portable = _js.load(file, **(json_settings or {}), )
        return klass.from_portable(portable, **kwargs, )

    #]




"""
"""


#[

from __future__ import annotations

import numpy as _np
import scipy as _sp

from .. import equations as _equations
from .. import quantities as _quantities
from ..incidences import main as _incidences
from ..aldi.differentiators import (AtomFactoryProtocol, Context, )
from ..aldi import maps as _maps
from ..aldi.maps import (ArrayMap, )

from typing import (TYPE_CHECKING, )
if TYPE_CHECKING:
    from typing import (Any, Protocol, Callable, )
    from collections.abc import (Iterable, )

#]


class _Jacobian:
    """
    """
    #[

    _atom_factory: AtomFactoryProtocol = ...

    def _populate_map(self, *args, **kwargs, ) -> None:
        raise NotImplementedError

    def eval(self, *args, **kwargs, ) -> _np.ndarray:
        raise NotImplementedError

    def __init__(
        self,
        equations: Iterable[_equations.Equation],
        wrt_something: Iterable[Any],
        qid_to_logly: dict[int, bool],
        /,
        *,
        context: dict[str, Callable] | None = None,
        initial: _np.ndarray | None = None,
        columns_to_eval: Iterable[int] | None = None,
        num_columns_to_eval: int = 1,
        **kwargs,
    ) -> None:
        """
        """
        self._columns_to_eval = _np.array(columns_to_eval, dtype=int, ) if columns_to_eval is not None else None
        self._num_columns_to_eval = num_columns_to_eval if columns_to_eval is None else len(columns_to_eval)
        #
        # Extract eids from equations
        eids = tuple(eqn.id for eqn in equations)
        wrt_something = tuple(wrt_something)
        #
        # Collect w.r.t. items (tokens, qids) which each equation is to be
        # differentiated
        eid_to_wrts = self._create_eid_to_wrts(equations, wrt_something, )
        #
        # Create the map from eids to rhs offsets; the offset is the number
        # of rows in the Jacobian matrix that precede the equation
        eid_to_rhs_offset = _maps.create_eid_to_rhs_offset(eids, eid_to_wrts, )
        #
        self._populate_shape(eids, wrt_something, )
        self._populate_map(
            eids,
            eid_to_wrts,
            wrt_something,
            eid_to_rhs_offset,
            rhs_column=0,
            lhs_column_offset=0,
        )
        #
        self._aldi_context = Context(
            self._atom_factory,
            equations,
            eid_to_wrts=eid_to_wrts,
            qid_to_logly=qid_to_logly,
            context=context,
        )
        #
        self.sparse_pattern = None
        self._matrix = None

    def _populate_shape(
        self,
        eids: Collection[int],
        wrt_something: Collection[Any],
    ) -> None:
        """
        """
        self._shape = len(eids)*self._num_columns_to_eval, len(wrt_something),

    def _create_eid_to_wrts(
        self,
        equations: Iterable[_equations.Equation],
        wrts: Iterable[Any],
        /,
    ) -> dict[int, tuple[Any, ...]]:
        """
        """
        raise NotImplementedError

    def _create_jacobian_matrix(self, diff_array, /, ) -> Any:
        """
        Create Jacobian as scipy sparse matrix
        """
        jacobian_matrix = self._initialize_jacobian_matrix()
        jacobian_matrix[self._map.lhs] = diff_array[self._map.rhs]
        return jacobian_matrix

    #]


class DenseJacobian(_Jacobian):
    """
    """
    #[

    def _initialize_jacobian_matrix(self, /, ) -> _np.ndarray:
        """
        Initialize the Jacobian matrix as a dense numpy array
        """
        return _np.zeros(self._shape, dtype=float, )

    #]


class SparseJacobian(_Jacobian):
    """
    """
    #[

    def __init__(self, *args, terminator: Terminator | None, **kwargs, ) -> None:
        """
        """
        super().__init__(*args, **kwargs, )
        self.sparse_pattern = self._map.lhs

    def _initialize_jacobian_matrix(self, /, ) -> _sp.sparse.csc_matrix:
        """
        Initialize the Jacobian matrix as a sparse scipy array
        """
        num_entries = len(self.sparse_pattern[0])
        return _sp.sparse.csc_matrix(
            ([0]*num_entries, self.sparse_pattern, ),
            shape=self._shape, dtype=float,
        )
    #]




"""
"""


#[
from __future__ import annotations

import enum as _en
#]


class Flags(_en.IntFlag, ):
    """
    """
    #[

    DEFAULT = 0
    LINEAR = _en.auto()
    FLAT = _en.auto()
    DETERMINISTIC = _en.auto()

    @property
    def is_linear(self, /, ) -> bool:
        return Flags.LINEAR in self

    @property
    def is_nonlinear(self, /, ) -> bool:
        return not self.is_linear

    @property
    def is_flat(self, /, ) -> bool:
        return Flags.FLAT in self

    @property
    def is_nonflat(self, /, ) -> bool:
        return not self.is_flat

    @property
    def is_deterministic(self, /, ) -> bool:
        return Flags.DETERMINISTIC in self

    @property
    def is_stochastic(self, /, ) -> bool:
        return not self.is_deterministic

    def update_from_kwargs(self, /, **kwargs) -> Self:
        linear = kwargs.get("linear") if kwargs.get("linear") is not None else self.is_linear
        flat = kwargs.get("flat") if kwargs.get("flat") is not None else self.is_flat
        deterministic = kwargs.get("deterministic") if kwargs.get("deterministic") is not None else self.is_deterministic
        return type(self).from_kwargs(linear=linear, flat=flat, deterministic=deterministic, )

    @classmethod
    def from_kwargs(cls: type, **kwargs, ) -> Self:
        self = cls.DEFAULT
        if kwargs.get("linear") or kwargs.get("is_linear"):
            self |= cls.LINEAR
        if kwargs.get("flat") or kwargs.get("is_flat"):
            self |= cls.FLAT
        if kwargs.get("deterministic") or kwargs.get("is_deterministic"):
            self |= cls.DETERMINISTIC
        return self

    def to_portable(self, /, ) -> dict[str, bool]:
        """
        """
        return {
            "is_linear": self.is_linear,
            "is_flat": self.is_flat,
            "is_deterministic": self.is_deterministic,
        }

    @classmethod
    def from_portable(klass, portable: dict[str, bool], ) -> Self:
        return klass.from_kwargs(**portable, )

    #]




"""
"""


#[

from __future__ import annotations

import math as _ma
from typing import Literal
from numbers import Real
from collections.abc import Iterable, Sequence
import plotly.graph_objects as _pg
import plotly.subplots as _ps
import plotly.io as _pi
import documark as _dm

from ..dates import Period, Frequency

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Any
#]


__all__ = (
    "add_line",
    "add_scatter",
    "add_bar",
    "add_histogram",
    "customize_tick_labels",
    "add_trace",
    "add_highlight",
    "add_vline",
    "auto_tiles",
    "freeze_span",
    "highlight",
    "make_subplots",
    "get_date_axis_mode",
    "resolve_date_format",
    "resolve_subplot_reference",
    "set_default_plotly_renderer",
    "update_subplot_title",
    "vline",
    "PlotlyDateAxisModeType",
)


PlotlyDateAxisModeType = Literal["instant", "period", ]


_DATE_FORMAT_STYLES = {
    "sdmx": {
        Frequency.YEARLY: "%Y",
        Frequency.HALFYEARLY: "%Y-%m",
        Frequency.QUARTERLY: "%Y-Q%q",
        Frequency.MONTHLY: "%Y-%m",
        Frequency.WEEKLY: "%Y-W%W",
        Frequency.DAILY: "%Y-%m-%d",
        Frequency.INTEGER: None,
    },
    "compact": {
        Frequency.YEARLY: "%yY",
        Frequency.HALFYEARLY: "%yM%m",
        Frequency.QUARTERLY: "%yQ%q",
        Frequency.MONTHLY: "%yM%m",
        Frequency.WEEKLY: "%yW%W",
        Frequency.DAILY: "%y%b%d",
        Frequency.INTEGER: None,
    },
}


def resolve_date_format(
    date_format_style: str | dict[int, str] | None,
    frequency: Frequency,
) -> str:
    """
    """
    if not date_format_style:
        date_format_style = "sdmx"
    if isinstance(date_format_style, str):
        date_format_style = _DATE_FORMAT_STYLES[date_format_style]
    return date_format_style[frequency]


def set_default_plotly_renderer(renderer: str, ) -> None:
    """
    """
    _pi.renderers.default = renderer


_EMPTY_SUBPLOT_TITLE = " "
_DEFAULT_VERTICAL_SPACING = 0.1
_DEFAULT_HORIZONTAL_SPACING = 0.05


@_dm.reference(category="arrange", )
def make_subplots(
    rows_columns: Iterable[int] | None = None,
    subplot_titles: Sequence[str] | bool | None = True,
    vertical_spacing: float | None = None,
    horizontal_spacing: float | None = None,
    figure_height: int | None = None,
    figure_title: str | None = None,
    show_legend: bool = False,
    **kwargs,
) -> _pg.Figure:
    """
················································································

==Create a figure with multiple subplots==

················································································
    """
    if vertical_spacing is None:
        vertical_spacing = _DEFAULT_VERTICAL_SPACING
    if horizontal_spacing is None:
        horizontal_spacing = _DEFAULT_HORIZONTAL_SPACING
    num_rows = rows_columns[0] if rows_columns else 1
    num_columns = rows_columns[1] if rows_columns else 1
    total_num_subplots = num_rows * num_columns
    if subplot_titles is True:
        subplot_titles = (_EMPTY_SUBPLOT_TITLE, ) * total_num_subplots
    figure = _ps.make_subplots(
        rows=num_rows,
        cols=num_columns,
        subplot_titles=subplot_titles or None,
        vertical_spacing=vertical_spacing,
        horizontal_spacing=horizontal_spacing,
        **kwargs,
    )
    #
    if figure_title is not None:
        figure.update_layout(title=figure_title, )
    #
    if figure_height is not None:
        figure.update_layout(height=figure_height, )
    #
    figure.update_layout(showlegend=show_legend, )
    #
    return figure

@_dm.reference(category="custom", )
def add_highlight(
    figure: _pg.Figure,
    span: Iterable[Period] | None = None,
    subplot: tuple[int, int] | int | None = None,
    color: str = "rgba(0, 0, 0, 0.15)",
) -> _pg.Figure:
    """
················································································

==Highlight a certain date span in charts==

Highlight a certain time span in a figure. The span can be specified either as a
time `Span`, or a tuple of two `Periods`.

    irispie.plotly.highlight(span)

················································································
    """
    highlight_start, highlight_end = _resolve_highlight_span(figure, span, )
    _, index = resolve_subplot_reference(figure, subplot, )
    xref = f"x{index+1}" if index else "x"
    yref = f"y{index+1} domain" if index else "y domain"
    shape = {
        "type": "rect",
        "xref": xref,
        "x0": highlight_start,
        "x1": highlight_end,
        "yref": yref,
        "y0": 0,
        "y1": 1,
        "fillcolor": color,
        "line": {"width": 0, },
    }
    figure.add_shape(shape, )


highlight = add_highlight


def add_vline(
    figure,
    position,
    subplot: tuple[int, int] | int | None = None,
    color: str = "rgba(0, 0, 0, 0.5)",
    width: int = None,
) -> None:
    r"""
    """
    date_axis_mode = get_date_axis_mode(figure, )
    plotly_row_col, _ = resolve_subplot_reference(figure, subplot, )
    if hasattr(position, "to_plotly_date"):
        position = position.to_plotly_date(mode=date_axis_mode, )
    figure.add_vline(position, **plotly_row_col, line={"color": color, "width": width, }, )


vline = add_vline


def freeze_span(
    figure: _pg.Figure,
    span: Iterable[Period],
    subplot: tuple[int, int] | int | None = None,
) -> None:
    """
    """
    start, end = _resolve_figure_span(figure, span, )
    xaxis_update = {
        "range": (start, end),
        "autorange": False,
    }
    plotly_row_col, *_ = resolve_subplot_reference(figure, subplot, )
    figure.update_xaxes(xaxis_update, **plotly_row_col, )


def _resolve_figure_span(
    figure: _pg.Figure,
    span: Iterable[Period] | None = None,
) -> tuple[str, str]:
    """
    """
    #[
    if hasattr(span, "start", ) and hasattr(span, "end", ):
        start = span.start if span.start else None
        end = span.end if span.end else None
    elif span:
        span = tuple(span, )
        start, end = span[0], span[-1]
    else:
        start, end = None, None
    figure_span = figure.layout.xaxis.range
    mode = get_date_axis_mode(figure, ) or "period"
    start = start.to_plotly_edge_before(mode=mode, ) if start is not None else figure_span[0]
    end = end.to_plotly_edge_after(mode=mode, ) if end is not None else figure_span[-1]
    return start, end,
    #]


def _resolve_highlight_span(
    figure: _pg.Figure,
    span: Iterable[Period] | None = None,
) -> tuple[str, str]:
    """
    """
    #[
    if hasattr(span, "start", ) and hasattr(span, "end", ):
        start = span.start if span.start else None
        end = span.end if span.end else None
    elif span:
        span = tuple(span, )
        start, end = span[0], span[-1]
    else:
        start, end = None, None
    figure_span = figure.layout.xaxis.range
    mode = get_date_axis_mode(figure, )
    highlight_start = start.to_plotly_edge_before(mode=mode, ) if start is not None else figure_span[0]
    highlight_end = end.to_plotly_edge_after(mode=mode, ) if end is not None else figure_span[-1]
    return highlight_start, highlight_end,
    #]


def resolve_subplot_reference(
    figure: _pg.Figure,
    subplot: tuple[int, int] | int | None,
) -> tuple[dict[str, int], int]:
    r"""
    """
    if subplot is None:
        return {}, None
    rows, columns = figure._get_subplot_rows_columns()
    num_rows = len(rows)
    num_columns = len(columns)
    num_tiles = num_rows * num_columns
    if isinstance(subplot, Sequence, ):
        row, column, = subplot
        index = row * num_columns + column
        plotly_row_col = {"row": row+1, "col": column+1, }
        return plotly_row_col, index,
    if isinstance(subplot, Real, ):
        index = int(subplot, ) if subplot >= 0 else num_tiles + int(subplot)
        row = index // num_columns
        column = index % num_columns
        plotly_row_col = {"row": row+1, "col": column+1, }
        return plotly_row_col, index,
    raise TypeError(f"Invalid subplot reference: {subplot}")


def add_trace(
    figure: _pg.Figure,
    trace,
    subplot: tuple[int, int] | int | None = None,
) -> None:
    plotly_row_col, _ = resolve_subplot_reference(figure, subplot, )
    figure.add_trace(trace, **plotly_row_col, )


def customize_tick_labels(
    figure: _pg.Figure,
    tick_labels: Iterable[str],
    tick_values: Iterable[Any],
    subplot: tuple[int, int] | int | None = None,
) -> None:
    plotly_row_col, _ = resolve_subplot_reference(figure, subplot, )
    figure.update_xaxes(
        ticktext=list(tick_labels),
        tickvals=list(tick_values),
        **plotly_row_col,
    )

def update_subplot_title(
    figure: _pg.Figure,
    subplot: tuple[int, int] | int,
    title: str,
) -> None:
    r"""
    """
    _, index = resolve_subplot_reference(figure, subplot, )
    annotation = next(figure.select_annotations(index, ), None, )
    if annotation:
        annotation.text = title


def get_date_axis_mode(figure: _pg.Figure, ) -> PlotlyDateAxisModeType | None:
    r"""
    """
    #[
    xaxis = figure.layout.xaxis
    return xaxis.ticklabelmode if xaxis.type == "date" else None
    #]


def add_histogram(figure, *args, subplot=None, **kwargs, ) -> _pg.Histogram:
    r"""
    """
    #[
    trace = _pg.Histogram(*args, **kwargs, )
    add_trace(figure, trace, subplot=subplot, )
    return trace
    #]


def add_bar(figure, *args, subplot=None, **kwargs, ) -> _pg.Histogram:
    r"""
    """
    #[
    trace = _pg.Bar(*args, **kwargs, )
    add_trace(figure, trace, subplot=subplot, )
    return trace
    #]


def add_line(figure, *args, subplot=None, **kwargs, ) -> _pg.Scatter:
    r"""
    """
    #[
    mode = kwargs.pop("mode", "lines+markers", )
    trace = _pg.Scatter(*args, **kwargs, mode=mode, )
    add_trace(figure, trace, subplot=subplot, )
    return trace


def add_scatter(figure, *args, subplot=None, **kwargs, ) -> _pg.Scatter:
    r"""
    """
    #[
    mode = kwargs.pop("mode", "markers", )
    trace = _pg.Scatter(*args, **kwargs, mode=mode, )
    add_trace(figure, trace, subplot=subplot, )
    return trace
    #]


def auto_tiles(num_charts, ) -> tuple[int, int]:
    r"""
    """
    #[
    n = _ma.ceil(_ma.sqrt(num_charts, ), )
    if n * (n-1) >= num_charts:
        return (n, n-1, )
    return (n, n, )
    #]


def auto_tiles(num_charts, ) -> tuple[int, int]:
    r"""
    """
    #[
    n = _ma.ceil(_ma.sqrt(num_charts, ), )
    if n * (n-1) >= num_charts:
        return (n, n-1, )
    return (n, n, )
    #]


# @_dm.reference(
#     path=("visualization_reporting", "ez_plotlyper.md", ),
#     categories={
#         "arrange": "Arranging charts",
#         "custom": "Customizing charts",
#     },
# )
# class plotly:
#     """
# ················································································
# 
# Plotly wrapper
# ===============
# 
# ················································································
#     """
#     #[
# 
#     Figure = _pg.Figure
#     make_subplots = staticmethod(make_subplots)
#     update_subplot_title = staticmethod(update_subplot_title)
#     resolve_subplot_reference = staticmethod(resolve_subplot_reference)
#     highlight = staticmethod(highlight)
#     freeze_span = staticmethod(freeze_span)
#
#    #]




r"""
First-order stacked-time system
"""

#[

from __future__ import annotations

import numpy as _np
import scipy as _sp

from .. import has_variants as _has_variants
from ..simultaneous.main import Simultaneous
from .. import quantities as _quantities
from ..quantities import QuantityKind, Quantity
from ..databoxes.main import Databox
from ..dataslates.main import Dataslate
from .. import dates as _dates
from ..dates import Period

from ._invariants import Invariant
from ._variants import Variant
from ..fords import covariances as _covariances
from . import _slatable_protocols as _slatable_protocols

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from collections.abc import Iterable
    from ..simultaneous.main import Simultaneous

#]


__all__ = [
    "Stacker",
]


MarginalTupe = tuple[_np.ndarray, _np.ndarray]


class Stacker(
    _slatable_protocols.Inlay,
    _has_variants.Mixin,
    _quantities.Mixin,
):
    r"""
    """
    #[

    __slots__ = (
        "_invariant",
        "_variants",
    )

    @classmethod
    def from_simultaneous(
        klass,
        model: Simultaneous,
        span: Iterable[Period],
        **kwargs,
    ) -> Self:
        """
        """
        self = klass()
        self._invariant = Invariant.from_simultaneous(model, span, **kwargs, )
        zipped_iters = zip(
            model.iter_solution(),
            model.iter_std_name_to_value(),
            model.iter_cov_u(),
        )
        self._variants = [
            Variant.from_solution_and_stds(self._invariant, solution, std_name_to_value, cov_u, )
            for solution, std_name_to_value, cov_u in zipped_iters
        ]
        return self

    @property
    def max_lag(self, ) -> int:
        return self._invariant.max_lag

    @property
    def max_lead(self, ) -> int:
        return self._invariant.max_lead

    @property
    def base_periods(self, ) -> Iterable[Period]:
        return self._invariant.base_periods

    @property
    def stacked_vector(self, ) -> tuple[str]:
        r"""
        """
        names = self.transition_variable_names
        periods = self.base_periods
        return tuple( f"{n}[{p}]" for p in periods for n in names )

    @property
    def transition_variable_names(self, ) -> tuple[str]:
        r"""
        """
        qid_to_name = self.create_qid_to_name()
        tokens = self._invariant.stacked_solution_vectors.transition_variables
        return tuple( qid_to_name[t.qid] for t in tokens )

    def create_dataslate(
        self,
        databox: Databox,
        #
        num_variants: int | None = None,
        shocks_from_data: bool = False,
        stds_from_data: bool = False,
    ) -> Dataslate:
        r"""
        """
        num_variants = self.resolve_num_variants_in_context(num_variants, )
        slatable = self.slatable_for_marginal(
            shocks_from_data=shocks_from_data,
            stds_from_data=stds_from_data,
        )
        return Dataslate.from_databox_for_slatable(
            slatable, databox, self.base_periods,
            num_variants=num_variants,
        )


    def calculate_marginal(
        self,
        databox: Databox,
        unpack_singleton: bool = True,
        **kwargs,
    ) -> MarginalType | list[MarginalType]:
        r"""
        """
        marginals = [
            self._calculate_marginal_for_variant(v, databox, **kwargs, )
            for v in self._variants
        ]
        return _has_variants.unpack_singleton(
            marginals, self.is_singleton,
            unpack_singleton=unpack_singleton,
        )

    def _calculate_marginal_for_variant(
        self,
        variant: Variant,
        databox: Databox,
        **kwargs,
    ) -> tuple[_np.ndarray, _np.ndarray]:
        r"""
        """
        #
        dataslate = self.create_dataslate(databox, num_variants=1, **kwargs, )
        dataslate.logarithmize()
        data_array = dataslate.get_data_variant()
        #
        y1, u0, v0, w0 = self._extract_data_arrays(data_array, variant, )
        #
        alpha0 = variant.init_med
        cov_alpha0 = variant.init_mse
        U = variant.U
        #
        cov_u = self._construct_cov_u(data_array, )
        cov_w = self._construct_cov_w(data_array, )
        #
        TT = variant.TT
        PP = variant.PP
        RR = variant.RR
        KK = variant.KK
        #
        y1_index = _where_observations(y1, )
        y1 = y1[y1_index]
        AA = variant.AA[y1_index, :]
        BB = variant.BB[y1_index, :]
        CC = variant.CC[y1_index, :]
        DD = variant.DD[y1_index]
        HH = variant.HH[y1_index, :]
        #
        num_alpha = alpha0.size
        num_u = u0.size
        num_w = w0.size
        ABH = _np.hstack((AA, BB, HH, ))
        TP = _np.hstack((TT, PP, ))
        auw0 = _np.hstack((alpha0, u0, w0, ))
        cov_auw0 = _sp.linalg.block_diag(cov_alpha0, cov_u, cov_w, )
        #
        # Nonstochastic part of x and y
        x_nonstochastic = RR @ v0 + KK
        y_nonstochastic = CC @ v0 + DD
        #
        # Unconditional distribution
        y0 = AA @ alpha0 + BB @ u0 + HH @ w0 + y_nonstochastic
        FF = AA @ cov_alpha0 @ AA.T + BB @ cov_u @ BB.T + HH @ cov_w @ HH.T
        FF = _covariances.symmetrize(FF)
        #
        # Prediction error
        pe = y1 - y0
        #
        # Update conditional on y
        # auw_update = cov_auw0 @ ABH.T @ _np.linalg.solve(FF, pe, )
        G = _right_div(cov_auw0 @ ABH.T, FF, )
        auw_update = G @ pe
        auw1 = auw0 + auw_update
        alpha1 = auw1[:num_alpha]
        au1 = auw1[:num_alpha+num_u]
        x1 = TP @ au1 + x_nonstochastic
        #
        # cov_auw_update = cov_auw0 @ ABH.T @ _np.linalg.solve(FF, ABH @ cov_auw0, )
        cov_auw_update = G @ ABH @ cov_auw0
        cov_auw1 = cov_auw0 - cov_auw_update
        cov_auw1 = _covariances.symmetrize(cov_auw1)
        #
        if variant.has_unknown_initial:
            MM = variant.MM[y1_index, :]
            delta = self._estimate_unknown_fixed_initial(FF, MM, pe, )
            auw1_diff = self._correct_for_unknown_fixed_initial(delta, variant.Xi, AA, G, )
            auw1 = auw1 + auw1_diff
        #
        # alpha1 = auw1[:num_alpha]
        # u1 = auw1[num_alpha:num_alpha+num_u]
        # w1 = auw1[num_alpha+num_u:]
        #
        au1 = auw1[:num_alpha+num_u]
        cov_au1 = cov_auw1[:num_alpha+num_u, :num_alpha+num_u]
        #
        mean_x1 = TP @ au1 + x_nonstochastic
        cov_x1 = TP @ cov_au1 @ TP.T
        return mean_x1, cov_x1,

    def _extract_data_arrays(
        self,
        data_array: _np.ndarray,
        variant: Variant,
    ) -> tuple[_np.ndarray, _np.ndarray, _np.ndarray, _np.ndarray, ]:
        r"""
        """
        s = self._invariant.squid
        u0 = data_array[s.u_qids, :].flatten(order="F", )
        v0 = data_array[s.v_qids, :].flatten(order="F", )
        w0 = data_array[s.w_qids, :].flatten(order="F", )
        y1 = data_array[s.y_qids, :].flatten(order="F", )
        return y1, u0, v0, w0,

    def _estimate_unknown_fixed_initial(
        self,
        FF: _np.ndarray,
        MM: _np.ndarray,
        pe: _np.ndarray,
    ) -> _np.ndarray:
        r"""
        """
        Fi_M = _left_div(FF, MM, )
        Mt_Fi_M = MM.T @ Fi_M
        Mt_Fi_pe = Fi_M.T @ pe
        delta, *_ = _np.linalg.lstsq(Mt_Fi_M, Mt_Fi_pe, rcond=None, )
        return delta

    def _correct_for_unknown_fixed_initial(
        self,
        delta: _np.ndarray,
        Xi: _np.ndarray,
        AA: _np.ndarray,
        G: _np.ndarray,
    ) -> Variant:
        r"""
        Calculate the correction to the state vector caused by the unknown fixed initial condition
        """
        # *_corr is the corrected state
        # *_diff is the difference between the corrected and uncorrected state
        #
        # alpha0_diff := alpha_corr - alpha0
        # alpha0_corr := alpha0 + Xi @ delta
        # hence alpha_diff := Xi @ delta
        alpha0_diff = Xi @ delta
        num_alpha = alpha0_diff.size
        num_auw = G.shape[0]
        num_uw = num_auw - num_alpha
        auw0_diff = _np.hstack((alpha0_diff, _np.zeros(num_uw, ), ), )
        y0_diff = AA @ alpha0_diff
        # pe = y1 - y0
        # pe_corr = y1 - y0_corr
        # pe_diff = pe_corr - pe = (y1 - y0_corr) - (y1 - y0) = y0 - y0_corr = -y0_diff
        pe_diff = -y0_diff
        # auw1 = auw0_corr + cov_auw0 @ ABH.T @ _np.linalg.solve(FF, pe_corr, )
        auw1_diff = auw0_diff + G @ pe_diff
        return auw1_diff

    def _construct_cov_u(
        self,
        data_array: _np.ndarray,
    ) -> _np.ndarray:
        """
        """
        std_u_qids = self._invariant.squid.std_u_qids
        std_u = data_array[std_u_qids, :].flatten(order="F", )
        return _np.diag(std_u**2)

    def _construct_cov_w(
        self,
        data_array: _np.ndarray,
    ) -> _np.ndarray:
        """
        """
        std_w_qids = self._invariant.squid.std_w_qids
        std_w = data_array[std_w_qids, :].flatten(order="F", )
        return _np.diag(std_w**2)

    #]


_left_div = _np.linalg.solve

def _right_div(
    A: _np.ndarray,
    B: _np.ndarray,
) -> _np.ndarray:
    r"""
    A @ inv(B)
    """
    return _left_div(B.T, A.T, ).T


def _where_observations(vector: _np.ndarray, ) -> list[int]:
    return _np.where(_np.isfinite(vector), )[0].tolist()

